"""
Measures of syncopation from the literature.

<small>**Author**: Mark Gotham</small>
"""

__author__ = "Mark Gotham"

from fractions import Fraction
from typing import Optional

from partitura import load_score


class SyncopationMetric:
    def __init__(self, path_to_score: Optional[str] = None):
        """
        The methods of this class implement syncopation metrics from the
        literature. These are typically based on simple data (note start
        times and similar).

        The parameters of this class allow users to run from a score
        (with onsets etc. deduced from there) or directly on their own
        data (the necessary parameters differ slightly for each method).

        Parameters
        ----------
        path_to_score:
            Path to the score in any supported format (e.g., MusicXML).
            Deduce any necessary onsets, beats etc. from the score as
            calculated by Partitura.
            Warning: Partitura takes “beats” from time signatures
            denominators, e.g., 6/8 has 6 “beats” (not 2).
        """
        self.path_to_score = path_to_score
        # TODO. TBC. May be redundant / better handled on a per-metric basis:
        self.note_array = None

    def load_note_array_from_score(self):
        """
        Parse a score and return Partitura's `.note_array()` with `include_metrical_position=True`.

        This should cover the required information.
        The note array's fields includes several fields of which methods here
        use the following (in their words):

        * 'onset_beat': onset time of the note in beats
        * 'duration_beat': duration of the note in beats

        These values are called in the form `note_array["onset_beat"]`.

        """
        if self.note_array is not None:
            print("already retrieved, skipping")
            return
        if self.path_to_score is None:
            raise ValueError("No score provided.")
        else:
            score = load_score(self.path_to_score)
            self.note_array = score.note_array(include_metrical_position=True)

    def weighted_note_to_beat_distance(
        self, onset_beats: Optional[list] = None
    ) -> float:
        """

        TODO: WIP - does not currently replicate answers in the literature; further investigation to follow

        The weighted note-to-beat distance measure (WNBD)
        measures the distance between note starts and
        records the traversing of beats, and the distance to the nearest beat.

        The authors clarify that “notes are supposed to end where the
        next note starts”,
        so we're working with the inter-note interval (INI), rather
        than the duration.
        Note that there are one fewer INI values than notes.

        Among the limitations is the incomplete definition of “beat”
        and the agnostic view of metre:
        “By strong beats we just mean pulses.” (§3.4).

        Parameters
        ----------
        onset_beats:
            User supplied data for the onset time of each note expressed
            in beats. Optional.

        Returns
        -------
        WNBD value (float)

        Examples
        --------
        We use the example of the son clave
        (also available from the `meter.profiles` module),
        adapting to match presentation in the literature.

        >>> son = [1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0]
        >>> onset_beats = vector_to_onset_beat(vector=son, beat_unit_length=4)
        >>> sm = SyncopationMetric()
        >>> sm.weighted_note_to_beat_distance(onset_beats=onset_beats)
        Fraction(14, 5)

        >>> hesitation = [1, 0, 1, 0, 1, 0, 0, 1]
        >>> onset_beats = vector_to_onset_beat(vector=hesitation, beat_unit_length=4)
        >>> sm = SyncopationMetric()
        >>> sm.weighted_note_to_beat_distance(onset_beats=onset_beats)
        Fraction(1, 2)

        >>> from amads.music import example
        >>> test_xml_file = str(example.fullpath("musicxml/ex1.xml"))
        >>> sm = SyncopationMetric(path_to_score=test_xml_file)
        >>> sm.weighted_note_to_beat_distance()
        Fraction(4, 3)

        """
        # onset_beats is required for user-provided,
        if onset_beats is None:  # if not seek a score on the class
            if self.path_to_score is not None:
                self.load_note_array_from_score()
                onset_beats = [
                    Fraction(float(x["onset_beat"])) for x in self.note_array
                ]  # type: ignore
                # Sic, Fraction via first: Partitura uses np.float32 and
                #    Fractions do not accept that type.
                # TODO revisit class handling of this retrieval when
                #    more algos are in
            else:
                raise ValueError("No score or user values provided.")

        per_note_syncopation_values = []

        durations = [j - i for i, j in zip(onset_beats[:-1], onset_beats[1:])]
        # Sic, although Partitura note_array provides durations,
        #    we're using INI here.

        for i in range(len(durations)):
            onset = onset_beats[i]
            if int(onset) == onset:  # starts on a beat
                per_note_syncopation_values.append(0)
            else:
                duration = durations[i]
                this_beat_int = int(onset)  # NB round down
                if (
                    onset + duration <= this_beat_int + 1
                ):  # ends before or at e_{i+1}
                    numerator = 1
                elif (
                    onset + duration <= this_beat_int + 2
                ):  # ends before or at e_{i+2}
                    numerator = 2
                else:  # if onset + duration > this_beat_int + 2: # ends after e_{i+2}
                    numerator = 1

                distance_to_nearest_beat = abs(
                    round(onset) - Fraction(onset)
                )  # Fraction
                per_note_syncopation_values.append(
                    Fraction(numerator, distance_to_nearest_beat)
                )

        return sum(per_note_syncopation_values) / (
            len(per_note_syncopation_values) + 1
        )


def vector_to_onset_beat(vector: list, beat_unit_length: int = 2):
    """
    Map from a vector to onset beat data via `vector_to_multiset`.

    Examples
    --------
    >>> son = [1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0, 1]  # Final 1 for cycle rotation
    >>> vector_to_onset_beat(vector=son, beat_unit_length=4) # NB different beat value
    (Fraction(0, 1), Fraction(3, 4), Fraction(3, 2), Fraction(5, 2), Fraction(3, 1), Fraction(4, 1))

    """
    onsets = [i for i, count in enumerate(vector) for _ in range(count)]
    return tuple(Fraction(x, beat_unit_length) for x in onsets)


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    import doctest

    doctest.testmod()



import pytest

from amads.core.pitch import Pitch


def test_pitch_comparison():
    # B3 is lower than C4
    b_3 = Pitch(59, alt=0)
    c_4 = Pitch(60, alt=0)
    assert b_3 < c_4


def test_non_int_pitch():
    micro = Pitch(60.5)
    assert micro.key_num == 60.5
    assert micro.alt == 0.5


def test_meta_pitch():
    """Currently, Pitch accepts Pitch objects."""
    p = Pitch(60)
    Pitch(p)


def test_constructor():
    p = Pitch(60, alt=-1.4)  # since alt is invalid, spell key_num "C4"
    assert p.key_num == 60
    assert p.alt == 0
    assert p.name == "C"
    assert p.name_with_octave == "C4"
    # invalid integer alt, and pitch class is not diatonic
    p1 = Pitch(61, alt=0)
    assert p1.key_num == 61
    assert p1.alt == 1
    assert p1.name == "C#"
    assert p1.name_with_octave == "C#4"
    # favor Bb over A#
    p1 = Pitch(70, alt=1.4)  # invalid alt forces default
    assert p1.key_num == 70
    assert p1.alt == -1
    assert p1.name == "Bb"
    assert p1.name_with_octave == "Bb4"
    # negative microtone
    p2 = Pitch(59.5, alt=-0.5)
    assert p2.key_num == 59.5
    assert p2.alt == -0.5
    assert p2.name == "C?"
    assert p2.name_with_octave == "C?4"
    # Valid specification with negative microtone alteration
    p3 = Pitch(64.4, alt=-0.6)
    assert p3.key_num == 64.4
    assert p3.alt == pytest.approx(-0.6)
    assert p3.name == "F?"
    assert p3.name_with_octave == "F?4"
    # between E and F, invalid microtone alteration
    p4 = Pitch(64.7, alt=1.7)
    assert p4.key_num == 64.7
    assert p4.alt == pytest.approx(-0.3)
    assert p4.name == "F?"
    assert p4.name_with_octave == "F?4"
    # microtone alteration near F#
    p5 = Pitch(66.3, alt=0.8)  # invalid alt forces default
    assert p5.key_num == 66.3
    assert p5.alt == pytest.approx(-0.7)
    assert p5.name == "G?"
    assert p5.name_with_octave == "G?4"
    # None for pitch
    p6 = Pitch(None)
    assert p6.key_num is None
    assert p6.alt == 0
    assert p6.name == "unpitched"
    assert p6.name_with_octave == "unpitched"

    # test enharmonic
    p7 = Pitch("A5").enharmonic()
    assert p7.key_num == 81
    assert p7.alt == -2
    assert p7.name == "Bbb"
    assert p7.name_with_octave == "Bbb5"

    # test simplest_enharmonic
    p8 = Pitch("B##4").simplest_enharmonic(sharp_or_flat="sharp")
    assert p8.key_num == 73
    assert p8.alt == 1
    assert p8.name == "C#"
    assert p8.name_with_octave == "C#5"


def test_from_name():
    info = Pitch.from_name("C#4")
    assert info == (61, 1)
    info = Pitch.from_name("Db4")
    assert info == (61, -1)
    info = Pitch.from_name("")
    assert info == (60, 0)
    info = Pitch.from_name("", 4)
    assert info == (60, 0)
    info = Pitch.from_name("Gm", 3, "mp")
    assert info == (54, -1)
    info = Pitch.from_name("Ap3", 4, "mp")
    assert info == (58, 1)



# fmt: off
# flake8: noqa E129,E303
"""
Rudimentary `Chord` representation.
This is not intended to compete with fuller representations elsewhere
but to cover here in AMADS ("in house")
the basic functionality needed for the majority of use cases.

Note that the "chord" is almost always a somewhat abstract (often analytic) category,
even when included in notation.
Most symbols define limited elements.
The inclusion of pitch class is a normal minimum.
The root is common.
Spelling is often nominally present but with spelling as a more human-readable short-hand
(C# is clearer than 1)
with enharmonics actually treated arbitrarily (C# = Db).
Specific octaves are almost never specified.

The `Chord` here is defined in relation to the above as a broadly
pitch-class-level concept:
with a set of pitch classes,
and various options (root, quality label, and spelling).

This should explain the distinction of Chord from a `PitchCollection`:

Properties that cannot be determined from the available information are `None`.

We work with a quality registry (see `QUALITIES`) that maps canonical tonal qualities (mostly triads and sevenths)
to the corresponding semitone intervals above the root.
Further aliases normalise to those canonical names (see `_ALIASES`).

"""

import re
from typing import Optional, Sequence, Union

from amads.core.pitch import Pitch, PitchCollection
from amads.core.vector_transforms_checks import indices_to_indicator

QUALITIES: dict[str, tuple[int, ...]] = {
    "major":      (0, 4, 7),
    "minor":      (0, 3, 7),
    "diminished": (0, 3, 6),
    "augmented":  (0, 4, 8),
    "dominant7":  (0, 4, 7, 10),
    "major7":     (0, 4, 7, 11),
    "minor7":     (0, 3, 7, 10),
    "diminished7":(0, 3, 6, 9),
    "half-dim7":  (0, 3, 6, 10),
    "sus2":       (0, 2, 7),  # Note the rotation equivalence ...
    "sus4":       (0, 5, 7),  # ... between these two.
    "major6":     (0, 4, 7, 9),
    "minor6":     (0, 3, 7, 9),
    "dominant9":  (0, 4, 7, 10, 14),
    "major9":     (0, 4, 7, 11, 14),
    "minor9":     (0, 3, 7, 10, 14),
    "power":      (0, 7),  # "Power chord" = root and perfect 5th only. Very common in popular corpora.
    "root":       (0, ),  # Root only. Dubious as a "chord" but good to include e.g., used in Billboard corpus.
}


_ALIASES: dict[str, str] = {
    # triads
    "maj": "major", "M": "major", "": "major",
    "min": "minor", "m": "minor",
    "dim": "diminished", "°": "diminished",
    "aug": "augmented", "+": "augmented",
    # sevenths
    "7": "dominant7", "dom7": "dominant7",
    "maj7": "major7", "M7": "major7", "Δ": "major7", "Δ7": "major7",
    "min7": "minor7", "m7": "minor7",
    "dim7": "diminished7", "°7": "diminished7",
    "ø7": "half-dim7", "ø": "half-dim7", "m7b5": "half-dim7",
    # other
    "5": "power",
    "1": "root",
}

#: Reverse lookup: pitch-class intervals -> canonical quality name.
_INTERVAL_SET_TO_QUALITY: dict[frozenset[int], str] = {
    frozenset(intervals): name
    for name, intervals in QUALITIES.items()
}


# ---------------------------------------------------------------------------

_ROMAN_TO_DEGREE: dict[str, int] = {
    "I": 0, "II": 1, "III": 2, "IV": 3,
    "V": 4, "VI": 5, "VII": 6,
}

_ROMAN_RE = re.compile(
    r'^([b#]?)([IViv]+)([o°]?)(ø?)([\+]?)(\d*)$'
)

#: Scale intervals (minor = natural)
_MAJOR_SCALE = (0, 2, 4, 5, 7, 9, 11)
_MINOR_SCALE = (0, 2, 3, 5, 7, 8, 10)

# Diatonic triad qualities by root
_MAJOR_TRIADS = ("major", "minor", "minor", "major",
                 "major", "minor", "diminished")
_MINOR_TRIADS = ("minor", "diminished", "major", "minor",
                 "minor", "major",  "major")


def _canonical_quality(quality: str) -> str:
    """Resolve a quality string to a canonical name, raising ValueError if unknown."""
    q = quality.strip()
    if q in QUALITIES:
        return q
    if q in _ALIASES:
        return _ALIASES[q]
    raise ValueError(
        f"Unknown chord quality {quality!r}. "
        f"Known qualities: {sorted(QUALITIES)} and aliases: {sorted(_ALIASES)}"
    )


def _parse_root(root: Union[str, int, Pitch]) -> Pitch:
    """Return a pitch-class-level Pitch from a name, int, or Pitch."""
    if isinstance(root, Pitch):
        return Pitch(root.key_num % 12, root.alt)
    if isinstance(root, int):
        return Pitch(root % 12)
    return Pitch(root, octave=None)  # octave=None → pitch-class (key_num 0-11)


def _parse_key(key: Union[str, int, Pitch]) -> tuple[tuple[int, ...], int]:
    """Return (scale, tonic_pc) from a key string, int, or Pitch.

    Key convention:
    Explicit `"… major"` or `"… minor"` overrides anything else.
    With this, we look at the root name case:
    uppercase = major;
    lowercase = minor.

    """
    if isinstance(key, (int, Pitch)):
        return _MAJOR_SCALE, int(_parse_root(key).key_num) % 12
    key_str = str(key).strip()
    if key_str.lower().endswith(" minor"):
        scale, tonic_name = _MINOR_SCALE, key_str[:-6].strip().capitalize()
    elif key_str.lower().endswith(" major"):
        scale, tonic_name = _MAJOR_SCALE, key_str[:-6].strip().capitalize()
    elif key_str[0].islower():
        scale, tonic_name = _MINOR_SCALE, key_str.capitalize()
    else:
        scale, tonic_name = _MAJOR_SCALE, key_str
    return scale, int(_parse_root(tonic_name).key_num) % 12


# ---------------------------------------------------------------------------


class Chord:
    """
    Represents a chord as a set of pitch classes,
    with optional root, quality, and spelling.
    Properties that cannot be determined are set to `None`.

    Parameters
    ----------
    root : str | int | Pitch | None
        The root of the chord, as a pitch name ("C"), pitch-class int (0), or Pitch.
        If `None`, a rootless chord is created from `pitches`.
    quality : str | None
        Quality string, e.g. "major", "min7", "°"
        typically as imported from some user provided string, or corpus source.
        If both `root` and `quality` are given, `pitches` is derived automatically.
        If only `root` is given, quality remains `None`
        unless it can be inferred from `pitches`.
    pitches : sequence of (str | int | Pitch) | None
        Explicit pitch members.
        Each element may be a pitch name ("Bb"), a pitch-class int (10), or a Pitch.
        If given alongside `root` and `quality`, these are stored as the chord's spelling.
    bass : str | int | Pitch | None
        The bass note if it differs from the root (slash chords, inversions).

    Attributes
    ----------
    root : Pitch | None
    quality : str | None
        Canonical quality name, e.g. "major", "minor7".
    pitches : tuple[Pitch, ...] | None
        Spelling as pitch-class Pitches (octave == -1), in the order given.
    bass : Pitch | None
    key : str | None
        Set when constructed via :meth:`from_roman`.

    Examples
    --------
    These examples demonstrate different ways to build `Chord` objects, and some of the attributes.

    1. From a root + quality string:

    >>> c = Chord("C", "major")
    >>> c
    Chord(C major)

    >>> c == Chord("C", "maj")
    True

    Here are some derived properties:

    >>> c.pitch_class_set
    [0, 4, 7]

    >>> c.intervals
    (0, 4, 7)

    >>> c.pitch_class_vector
    (1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0)

    Here is a different quality exmaple (see `QUALTIES` for all options):

    >>> Chord("G", "dominant7").label
    'G7'

    2. From a Roman numeral, given a key.

    >>> Chord.from_roman("IV", "C").root.name
    'F'

    >>> Chord.from_roman("V7", "G").label
    'D7'

    Roman numerals are complex and not consistent across standards.
    This illustrates a case in point:
    The triad in a minor key on the second degree is diminshed diatonically,
    and the seventh is a half-diminshed.
    Parsing routines vary.
    Here we take "ii7" to imply the diatonic half-diminshed,
    "iiø7" is equivalent (not needed to specify the half-diminshed)
    and "iio7" modifies to diminished.
    See also the `from_scale_degree` method (also demo'd here).

    >>> Chord.from_roman("ii", "d").label
    'Eo'

    >>> Chord.from_roman("ii7", "d").label
    'Eø7'

    >>> Chord.from_roman("iiø7", "d").label
    'Eø7'

    >>> Chord.from_scale_degree(2, "d").label
    'Eo'

    Explicit suffix takes priority over scale-stacking
    >>> Chord.from_roman("iio7", "d").label
    'Eo7'

    3. From pitches directly, either as pitch name or pitch class:

    >>> Chord(pitches=["C", "Eb", "G"]).quality
    'minor'

    >>> Chord(pitches=[0, 3, 7]).quality
    'minor'

    >>> Chord(pitches=["C", "E", "G", "Bb"]).label
    'C7'
    """

    __slots__ = ["root", "quality", "pitches", "bass", "key"]

    root:    Optional[Pitch]
    quality: Optional[str]
    pitches: Optional[tuple[Pitch, ...]]
    bass:    Optional[Pitch]
    key:     Optional[str]

    def __init__(
        self,
        root:    Union[str, int, Pitch, None] = None,
        quality: Optional[str]                = None,
        pitches: Optional[Sequence]           = None,
        bass:    Union[str, int, Pitch, None] = None,
    ) -> None:

        self.key = None
        self.root = _parse_root(root) if root is not None else None
        self.quality = _canonical_quality(quality) if quality is not None else None
        if pitches is not None: # Pitches explicitly supplied:
            self.pitches = tuple(_parse_root(p) for p in pitches)
        elif self.root is not None and self.quality is not None: # Derive pitches from root + quality:
            intervals = QUALITIES[self.quality]
            root_pc   = self.root.key_num % 12
            self.pitches = tuple(
                Pitch((root_pc + i) % 12)
                for i in intervals
            )
        else:
            self.pitches = None

        # Derive quality from pitches:
        if self.quality is None and self.pitches is not None:
            self.quality = self._infer_quality()

        # Derive root from pitches + quality:
        if self.root is None and self.quality is not None and self.pitches is not None:
            self.root = self._infer_root()

        self.bass = _parse_root(bass) if bass is not None else None

    @classmethod
    def from_roman(cls, numeral: str, key: Union[str, int, Pitch], *, seventh: bool = False) -> "Chord":
        scale, tonic_pc = _parse_key(key)
        key_str = str(key).strip()  # preserved for chord.key attribute only
        raw = numeral.strip()

        m = _ROMAN_RE.match(raw)
        if not m:
            raise ValueError(f"Unrecognised Roman numeral {numeral!r}.")
        accidental, core_raw, dim_flag, hdim_flag, aug_flag, num_str = m.groups()

        core = core_raw.upper()
        if core not in _ROMAN_TO_DEGREE:
            raise ValueError(f"Unrecognised Roman numeral {numeral!r}.")

        add7 = bool(num_str) or seventh
        degree = _ROMAN_TO_DEGREE[core]

        # Handle chromatic roots (bVII, #IV etc.)
        root_pc = (tonic_pc + scale[degree]) % 12
        if accidental == 'b':
            root_pc = (root_pc - 1) % 12
        elif accidental == '#':
            root_pc = (root_pc + 1) % 12

        # Explicit suffix takes priority over scale-stacking
        if hdim_flag:
            quality = "half-dim7" if add7 else "diminished"  # ø alone = diminished triad conventionally
        elif dim_flag:  # o or °
            quality = "diminished7" if add7 else "diminished"
        elif aug_flag:
            quality = "augmented"
        else:
            # Derive quality by stacking scale tones
            n_tones = 4 if add7 else 3
            stacked_pcs = [
                (tonic_pc + scale[(degree + 2 * i) % 7]) % 12
                for i in range(n_tones)
            ]
            intervals = frozenset((pc - root_pc) % 12 for pc in stacked_pcs)
            quality = _INTERVAL_SET_TO_QUALITY.get(intervals)
            if quality is None:
                raise ValueError(
                    f"Scale-stacked intervals {intervals} for {numeral!r} in {key!r} "
                    f"don't match any known quality."
                )

        chord = cls(Pitch(root_pc), quality)
        chord.key = key_str
        return chord

    @classmethod
    def from_scale_degree(
            cls,
            degree: int,
            key: Union[str, int, Pitch],
            *,
            seventh: bool = False,
    ) -> "Chord":
        """
        Alternate constructor for making a `Chord` from a scale degree and key.

        Parameters
        ----------
        degree : int
            Scale degree.
            Note that this is 1-indexed (1 = tonic, ..., 7 = leading tone).
        key : str | int | Pitch
            Key, e.g. `"C"`, `"d"` (lower-case = minor) ...
        seventh : bool
            If True, stack the diatonic seventh above the triad.

        Returns
        -------
        Chord

        Raises
        ------
        ValueError
            If *degree* is outside 1–7.

        Examples
        --------
        >>> Chord.from_scale_degree(2, "C").label
        'Dm'

        >>> ii7 = Chord.from_scale_degree(2, "C", seventh=True)
        >>> ii7.label
        'Dm7'

        >>> ii7.intervals
        (0, 3, 7, 10)

        >>> Chord.from_scale_degree(7, "C").label
        'Bo'

        >>> Chord.from_scale_degree(2, "d").label
        'Eo'

        >>> Chord.from_scale_degree(2, "d", seventh=True).label
        'Eø7'
        """
        if not 1 <= degree <= 7:
            raise ValueError(f"Scale degree must be between 1 and 7, got {degree!r}.")
        numeral = ["I", "II", "III", "IV", "V", "VI", "VII"][degree - 1]
        return cls.from_roman(numeral, key, seventh=seventh)

    @classmethod
    def from_pitch_collection(cls, collection: PitchCollection) -> "Chord":
        """
        Alternate constructor for making a `Chord` from a PitchCollection.

        Quality is identified by matching the pitch-class set against the
        quality registry under all rotations (potential roots).  If exactly
        one root yields a known quality, both are set; if multiple match
        (e.g. augmented triads), root is left None.  Spelling is preserved
        from the collection.

        Parameters
        ----------
        collection : PitchCollection

        Returns
        -------
        Chord
        """
        pitches = collection.pitches
        pc_pitches: list[Pitch] = [Pitch(p.key_num % 12, p.alt) for p in pitches]
        pc_set = frozenset(p.key_num % 12 for p in pc_pitches)

        # try each member as potential root
        matches: list[tuple[Pitch, str]] = []
        for candidate_root in pc_set:
            intervals = tuple(sorted((pc - candidate_root) % 12 for pc in pc_set))
            key = frozenset(intervals)
            if key in _INTERVAL_SET_TO_QUALITY:
                matches.append((Pitch(candidate_root), _INTERVAL_SET_TO_QUALITY[key]))

        chord = cls.__new__(cls)
        chord.key  = None
        chord.bass = pc_pitches[0] if pc_pitches else None
        chord.pitches = tuple(pc_pitches)

        if len(matches) == 1:
            chord.root, chord.quality = matches[0]
        elif len(matches) > 1:
            # ambiguous (e.g. augmented): no single root
            chord.root    = None
            chord.quality = matches[0][1]  # at least quality is unambiguous
        else:
            chord.root    = None
            chord.quality = None

        return chord

    def _infer_quality(self) -> Optional[str]:
        """
        Helper to try and identify quality from self.pitches, trying each member as root."""
        if not self.pitches:
            return None
        pc_set = frozenset(int(p.key_num) % 12 for p in self.pitches)
        # try root first if known
        candidates = (
            [int(self.root.key_num) % 12] if self.root is not None
            else [int(p.key_num) % 12 for p in self.pitches]
        )
        for r in candidates:
            intervals = frozenset(sorted((pc - r) % 12 for pc in pc_set))
            if intervals in _INTERVAL_SET_TO_QUALITY:
                return _INTERVAL_SET_TO_QUALITY[intervals]
        return None

    def _infer_root(self) -> Optional[Pitch]:
        """
        Helper to find the root that makes self.quality consistent with self.pitches.
        """
        if not self.pitches or not self.quality:
            return None
        target = frozenset(QUALITIES[self.quality])
        for p in self.pitches:
            r = int(p.key_num) % 12
            intervals = frozenset((int(q.key_num) % 12 - r) % 12 for q in self.pitches)
            if intervals == target:
                return Pitch(r)
        return None

    @property
    def pitch_class_set(self) -> Optional[list[int]]:
        """Sorted list of pitch classes, or None if pitches unknown."""
        if self.pitches is None:
            return None
        return sorted(set(int(p.key_num) % 12 for p in self.pitches))

    @property
    def pitch_class_vector(self) -> Optional[tuple[int, ...]]:
        """12-dimensional indicator vector, or None if pitches unknown."""
        if self.pitch_class_set is None:
            return None
        return indices_to_indicator(self.pitch_class_set, indicator_length=12)

    @property
    def pitch_names(self) -> Optional[list[str]]:
        """Spelled pitch names (e.g. ["C", "Eb", "G"]), or None."""
        if self.pitches is None:
            return None
        return [p.name for p in self.pitches]

    def __contains__(self, item: Union[str, int, Pitch]) -> bool:
        pc = _parse_root(item).key_num % 12
        if self.pitches is not None:
            return any(int(p.key_num) % 12 == pc for p in self.pitches)
        if self.root is not None and self.quality is not None:
            root_pc = int(self.root.key_num) % 12
            return ((pc - root_pc) % 12) in QUALITIES[self.quality]
        return False

    def __len__(self) -> int:
        """Number of distinct pitch classes."""
        return len(self.pitch_class_set) if self.pitch_class_set is not None else 0

    # Hereafter the properties move to those usable when root is known

    @property
    def intervals(self) -> Optional[tuple[int, ...]]:
        """Semitone intervals above the root, or None if root unknown."""
        if self.root is None or self.pitches is None:
            return None
        r = int(self.root.key_num) % 12
        return tuple(sorted((int(p.key_num) % 12 - r) % 12 for p in self.pitches))

    @property
    def inversion(self) -> Optional[int]:
        """0 = root position, 1 = first inversion, etc.; None if undetermined."""
        if self.root is None or self.pitches is None or not self.pitches:
            return None
        bass_pc = int(self.pitches[0].key_num) % 12
        root_pc = int(self.root.key_num) % 12
        if bass_pc == root_pc:
            return 0
        if self.intervals is None:
            return None
        intervals_list = list(self.intervals)
        bass_interval = (bass_pc - root_pc) % 12
        try:
            return intervals_list.index(bass_interval)
        except ValueError:
            return None

    @property
    def label(self) -> Optional[str]:
        """
        Short label like "Cmaj7", "Dmin", "G7".
        Broadly matches leadsheet notation.

        Returns `None` if either or both of root or quality unknown.
        """
        if self.root is None or self.quality is None:
            return None
        root_name = self.root.name

        # Concise suffixes for common qualities
        _LABEL_SUFFIX: dict[str, str] = {
            "major":       "",
            "minor":       "m",
            "diminished":  "o",
            "augmented":   "+",
            "dominant7":   "7",
            "major7":      "M7",
            "minor7":      "m7",
            "diminished7": "o7",
            "half-dim7":   "ø7",
            "sus2":        "sus2",
            "sus4":        "sus4",
            "major6":      "6",
            "minor6":      "min6",
            "dominant9":   "9",
            "major9":      "maj9",
            "minor9":      "min9",
            "power":       "5",
        }
        suffix = _LABEL_SUFFIX.get(self.quality, self.quality)

        bass_str = ""
        if self.bass is not None:
            bass_pc = int(self.bass.key_num) % 12
            root_pc = int(self.root.key_num) % 12
            if bass_pc != root_pc:
                bass_str = f"/{self.bass.name}"

        return f"{root_name}{suffix}{bass_str}"

    def roman_numeral(self, key: Union[str, int, Pitch]) -> Optional[str]:
        """Return the Roman numeral of this chord in the given key, or None."""
        if self.root is None or self.quality is None:
            return None

        scale, tonic_pc = _parse_key(key)
        root_pc  = int(self.root.key_num) % 12
        interval = (root_pc - tonic_pc) % 12

        if interval not in scale:
            return None  # chromatic root — don't attempt diatonic numeral

        degree = scale.index(interval)
        numerals = ["I", "II", "III", "IV", "V", "VI", "VII"]
        numeral  = numerals[degree]

        # lowercase for minor/diminished
        if self.quality in ("minor", "minor7", "diminished", "diminished7", "half-dim7"):
            numeral = numeral.lower()
        if self.quality in ("diminished", "diminished7"):
            numeral += "°"
        if self.quality in ("dominant7", "major7", "minor7",
                            "diminished7", "half-dim7"):
            numeral += "7"

        return numeral

    # Dunder / display

    def __repr__(self) -> str:
        parts = []
        if self.root is not None:
            parts.append(self.root.name)
        if self.quality is not None:
            parts.append(self.quality)
        if not parts and self.pitches is not None:
            parts.append(str(self.pitch_names))
        inner = " ".join(parts) if parts else "unknown"
        return f"Chord({inner})"

    def __eq__(self, other: object) -> bool:
        """Enharmonic pitch-class equality: same pc set, root, quality."""
        if not isinstance(other, Chord):
            return False
        return (
            self.pitch_class_set == other.pitch_class_set
            and (
                (self.root is None and other.root is None)
                or (
                    self.root is not None
                    and other.root is not None
                    and self.root.key_num % 12 == other.root.key_num % 12
                )
            )
            and self.quality == other.quality
        )

    def __hash__(self) -> int:
        pcs = tuple(self.pitch_class_set or [])
        root_pc = int(self.root.key_num) % 12 if self.root is not None else None
        return hash((pcs, root_pc, self.quality))


# ------------------------------------------------------------------------

if __name__ == "__main__":
    import doctest

    doctest.testmod()



from collections import OrderedDict
from typing import List, Optional

from amads.core.basics import Score


class MelodyTokenizer:
    """Abstract base class for tokenizing melodies into n-grams."""

    def __init__(self):
        """Initialize the tokenizer."""
        self.ioi_data = {}  # Dictionary to store IOI information for notes

    def tokenize(self, score: Score) -> List[List]:
        """Tokenize a melody into phrases. (Unimplemented abstract method.)

        Parameters
        ----------
        score : Score
            A Score object containing a melody

        Returns
        -------
        list[list]
            List of tokenized phrases

        Raises
        ------
        NotImplementedError
            if this method of MelodyTokenizer is called
        """
        raise NotImplementedError


class FantasticTokenizer(MelodyTokenizer):
    """This tokenizer produces the M-Types as defined in the FANTASTIC toolbox [1].

    An M-Type is a sequence of musical symbols (pitch intervals and duration ratios)
    that represents a melodic fragment, similar to how an n-gram represents a sequence
    of n consecutive items from a text. The length of an M-Type can vary, just like
    n-grams can be of different lengths (bigrams, trigrams, etc.)

    The tokenizer takes a score as the input, and returns a dictionary of unique
    M-Type (n-gram) counts.

    Attributes
    ----------
    phrase_gap : float
        Time gap in seconds that defines phrase boundaries
    tokens : list
        List of tokens after tokenization

    References
    ----------
    [1] Müllensiefen, D. (2009). Fantastic: Feature ANalysis Technology Accessing
        STatistics (In a Corpus): Technical Report v1.5
    """

    def __init__(self):
        super().__init__()
        self.tokens = []

    def tokenize(self, score: Score) -> List:
        """Tokenize a melody into M-Types.

        Parameters
        ----------
        score : Score
            Score object containing melody to tokenize

        Raises
        ------
        ValueError
            if score has more than one part, if the part has concurrent
            notes (IOI == 0) or if a Note in the part has a tie.

        Returns
        -------
        list
            List of M-Type tokens
        """
        # Extract notes and calculate IOIs using get_notes
        notes = score.calc_differences(["ioi-ratio", "interval"])
        if len(notes) != 1:
            raise ValueError("score has more than one Part")
        notes = notes[0]
        tokens = []

        # Skip if phrase is too short
        if len(notes) < 2:
            return tokens

        for note in notes[1:]:
            pitch_interval = note.get("interval")
            pitch_interval_class: Optional[str] = self.classify_pitch_interval(
                pitch_interval
            )
            ioi_ratio_class = self.classify_ioi_ratio(note.get("ioi_ratio"))
            token = MType(pitch_interval_class, ioi_ratio_class)
            tokens.append(token)
        return tokens

    def classify_pitch_interval(
        self, pitch_interval: Optional[int]
    ) -> Optional[str]:
        """Classify pitch interval according to Fantastic's interval class scheme.

        Parameters
        ----------
        pitch_interval : int or None
            Interval in semitones between consecutive notes

        Returns
        -------
        Optional[str]
            Interval class label (e.g. 'd8', 'd7', 'u2', etc.)
            'd' = downward interval,
            'u' = upward interval,
            's' = same pitch, and
            't' = tritone.
            Returns None if input is None
        """
        # Clamp interval to [-12, 12] semitone range
        if pitch_interval is None:
            return None

        if pitch_interval < -12:
            pitch_interval = -12
        elif pitch_interval > 12:
            pitch_interval = 12

        # Map intervals to class labels based on Fantastic's scheme
        return self.interval_map[pitch_interval]

    interval_map: OrderedDict[Optional[int], Optional[str]] = OrderedDict(
        [
            (None, None),  # missing interval
            (-12, "d8"),  # Descending octave
            (-11, "d7"),  # Descending major seventh
            (-10, "d7"),  # Descending minor seventh
            (-9, "d6"),  # Descending major sixth
            (-8, "d6"),  # Descending minor sixth
            (-7, "d5"),  # Descending perfect fifth
            (-6, "dt"),  # Descending tritone
            (-5, "d4"),  # Descending perfect fourth
            (-4, "d3"),  # Descending major third
            (-3, "d3"),  # Descending minor third
            (-2, "d2"),  # Descending major second
            (-1, "d2"),  # Descending minor second
            (0, "s1"),  # Unison
            (1, "u2"),  # Ascending minor second
            (2, "u2"),  # Ascending major second
            (3, "u3"),  # Ascending minor third
            (4, "u3"),  # Ascending major third
            (5, "u4"),  # Ascending perfect fourth
            (6, "ut"),  # Ascending tritone
            (7, "u5"),  # Ascending perfect fifth
            (8, "u6"),  # Ascending minor sixth
            (9, "u6"),  # Ascending major sixth
            (10, "u7"),  # Ascending minor seventh
            (11, "u7"),  # Ascending major seventh
            (12, "u8"),  # Ascending octave
        ]
    )

    interval_classes = OrderedDict.fromkeys(interval_map.values())

    interval_class_codes = {
        string: integer
        for integer, string in enumerate(interval_classes.keys())
    }

    def classify_ioi_ratio(self, ioi_ratio: Optional[float]) -> Optional[str]:
        """Classify an IOI ratio into relative rhythm classes.

        Parameters
        ----------
        ioi_ratio : float or None
            Inter-onset interval ratio between consecutive notes

        Returns
        -------
        str or None
            'q' for quicker (<0.8119),
            'e' for equal (0.8119-1.4946), and
            'l' for longer (>1.4946).
            Returns None if input is None.
        """
        if ioi_ratio is None:
            return None
        elif ioi_ratio < 0.8118987:
            return "q"
        elif ioi_ratio < 1.4945858:
            return "e"
        else:
            return "l"

    ioi_ratio_classes = [None, "q", "e", "l"]
    ioi_ratio_class_codes = {
        None: 0,
        "q": 1,
        "e": 2,
        "l": 3,
    }


class MType:
    """A class for representing M-Types."""

    def __init__(
        self,
        pitch_interval_class: Optional[str],
        ioi_ratio_class: Optional[str],
    ):
        self.pitch_interval_class = pitch_interval_class
        self.ioi_ratio_class = ioi_ratio_class
        self.integer = self.encode()

    def encode(self) -> int:
        n_ioi_ratio_classes = len(FantasticTokenizer.ioi_ratio_classes)

        interval_class_code = FantasticTokenizer.interval_class_codes[
            self.pitch_interval_class
        ]
        ioi_ratio_class_code = FantasticTokenizer.ioi_ratio_class_codes[
            self.ioi_ratio_class
        ]

        # interval_class_code = 0, ioi_ratio_class_code = 0 -> 0
        # interval_class_code = 0, ioi_ratio_class_code = 1 -> 1
        # interval_class_code = 0, ioi_ratio_class_code = 2 -> 2
        # interval_class_code = 0, ioi_ratio_class_code = 3 -> 3
        # interval_class_code = 1, ioi_ratio_class_code = 0 -> 4
        # interval_class_code = 1, ioi_ratio_class_code = 1 -> 5
        # interval_class_code = 1, ioi_ratio_class_code = 2 -> 6
        # interval_class_code = 1, ioi_ratio_class_code = 3 -> 7
        # interval_class_code = 2, ioi_ratio_class_code = 0 -> 8
        # interval_class_code = 2, ioi_ratio_class_code = 1 -> 9
        # interval_class_code = 2, ioi_ratio_class_code = 2 -> 10
        # interval_class_code = 2, ioi_ratio_class_code = 3 -> 11
        return interval_class_code * n_ioi_ratio_classes + ioi_ratio_class_code

    def __hash__(self):
        """Return the hash based on the integer attribute."""
        return hash(self.integer)

    def __eq__(self, other):
        """Check equality based on the integer attribute."""
        if isinstance(other, MType):
            return self.integer == other.integer
        return False

    def __repr__(self):
        """Return a string representation of the MType."""
        return f"MType {self.integer}"



"""
This module implements various functions useful for analyzing swing in jazz
and other genres.

<small>**Author**: Huw Cheston (2025)</small>

References
----------
- Benadon, F. (2006). Slicing the Beat: Jazz Eighth-Notes as Expressive
    Microrhythm.
    *Ethnomusicology*, 50(1), 73–98. https://doi.org/10.2307/20174424

- Corcoran, C., & Frieler, K. (2021). Playing It Straight: Analyzing Jazz
    Soloists’ Swing Eighth-Note Distributions with the Weimar Jazz Database.
    *Music Perception*, 38(4), 372–385.
    https://doi.org/10.1525/mp.2021.38.4.372

"""

from math import log2 as log2_
from typing import Iterable

import numpy as np

__author__ = "Huw Cheston"

TINY = 1e-8
LOW_BUR, HIGH_BUR = (
    0.25,
    4.0,
)  # the range of BURs to keep as defined in Corcoran & Frieler (2021)


# flake8: noqa: W605
def beat_upbeat_ratio(
    beats: Iterable[float],
    upbeats: Iterable[float],
    log2: bool = False,
    bounded: bool = False,
    lower_bound: float = LOW_BUR,
    upper_bound: float = HIGH_BUR,
) -> list[float]:
    r"""
    Extracts beat-upbeat ratio (BUR) values from an array of beats and upbeats.

    The beat-upbeat ratio (BUR) is used to analyze the amount of “swing” in two
    consecutive eighth-note beat durations. It is calculated by dividing the
    duration of the first (“long”) eighth-note beat by the duration of the
    second (“short”) eighth-note beat. A BUR value of 2 represents “perfect”
    swing (e.g., a triplet quarter note followed by a triplet eighth note),
    while a BUR of 1 represents “even” eighth-note durations.

    The formula for BUR is:

    $\text{BUR} = \frac{t_{a,b} - t_{a}}{t_{b} - t_{a,b}}$, where

    $t_a$ is the beat at position $a$, $t_b$ is the beat at position $b$,
    and $t_{a,b}$ is the single upbeat between beats $a$ and $b$.

    The function takes two iterables of timestamps: `beats` and `upbeats`.
    Both lists should be unique, and missing values should not be present.
    The function returns an array of BUR values with a size of
    `len(beats) - 1`. If multiple `upbeats` are found between two
    consecutive `beats` or if no `upbeat` is found, the BUR for those
    beats will be omitted and the corresponding value will be `None`.

    Additionally, the function can calculate the $log_2$ of the
    BUR values, where a value of 1.0 corresponds to “triplet” swing.
    This can be enabled by setting `log2=True`. The values can also
    be filtered to remove outliers by setting `bounded=True`, with
    the default values for the boundaries coming from Corcoran &
    Frieler (2021).

    Parameters
    ----------
    beats : Iterable[float]
        An array of beat timestamps. Should not overlap with `upbeats`.
    upbeats : Iterable[float]
        An array of upbeat timestamps.
    log2 : bool, optional
        If True, computes the log base 2 of BUR values, as used in [2].
        Defaults to False.
    bounded : bool, optional
        If True, filters out BUR values outside the specified range.
        Defaults to False.
    lower_bound : float, optional
        Lower boundary for filtering BUR values. Defaults to 0.25
        ($log_2$ value of -2).
    upper_bound : float, optional
        Upper boundary for filtering BUR values. Defaults to 4.0 ($log_2$
        value of 2).

    Returns
    -------
    list[float]
        A list of the calculated BUR values.

    Examples
    --------
    >>> my_beats = [0., 1., 2., 3.]
    >>> my_upbeats = [0.5, 1.75, 2.2]
    >>> beat_upbeat_ratio(my_beats, my_upbeats)
    [1., 3., 0.25]

    >>> # Consecutive beats without a matching upbeat will be skipped.
    >>> my_beats = [0., 1., 2., 3.]
    >>> my_upbeats = [0.5, 2.2]
    >>> beat_upbeat_ratio(my_beats, my_upbeats)
    [1., None, 0.25]

    >>> # Consecutive beats with multiple matching upbeats will be skipped.
    >>> my_beats = [0., 1., 2., 3.]
    >>> my_upbeats = [0.5, 1.5, 1.75, 1.8, 2.2]
    >>> beat_upbeat_ratio(my_beats, my_upbeats)
    [1., None, 0.25]

    >>> # Filter out outlying values by setting `bounded=True`.
    >>> my_beats = [0., 1., 2., 3.]
    >>> my_upbeats = [0.5, 1.75, 2.99]
    >>> beat_upbeat_ratio(my_beats, my_upbeats, bounded=True)
    [1., 3., None]

    """

    # Parse beats and upbeats to an array, sorting as required
    beats = np.sort(np.array(beats))
    upbeats = np.sort(np.array(upbeats))
    # Validate inputs and raise an errors if required
    _validate_bur_inputs(beats, upbeats)
    # Match beats with upbeats and return a 2d array of shape
    #    [[beat1, upbeat, beat2], [beat2, upbeat, beat3]]
    matched = match_beats_and_upbeats(beats, upbeats)
    # Raise an error if we cannot find any matches between beats and upbeats
    if all([np.isnan(i) for i in matched[:, 1]]):
        raise ValueError(
            "No matches found between beats and upbeats, cannot calculate BUR"
        )
    # Calculate the BUR for upbeats between consecutive beats
    #
    # Use isitem() because Python 3.10 will print np.float64(...) instead of ...
    # when the value is a numpy float:
    burs = [
        _bur(b1, upbeat, b2).item() if not np.isnan(upbeat) else None
        for b1, upbeat, b2 in matched
    ]
    # Apply our filtering if required
    # Filter before log_2 transform to make things simpler
    if bounded:
        burs = [i if lower_bound < i < upper_bound else None for i in burs]
    # Express as base-2 log if required
    if log2:
        burs = [log2_(b) if b is not None else None for b in burs]
    return burs


def mean_bur(
    beats: Iterable[float], upbeats: Iterable[float], **kwargs
) -> float:
    """Calculates mean BUR (or $log_2$ BUR) given a list of beats and upbeats"""
    # We use nanmean here as we may have null values in cases where
    # multiple upbeats match with a single pair of beats, or where
    # no upbeats match with a beat.
    #     I think this makes sense to avoid the user having to chop a
    # large list into multiple sublists depending on the presence of nans.
    return float(np.nanmean(beat_upbeat_ratio(beats, upbeats, **kwargs)))


def std_bur(
    beats: Iterable[float], upbeats: Iterable[float], **kwargs
) -> float:
    """Calculates standard deviation BUR (or $log_2$ BUR) given a list of beats and upbeats"""
    return np.nanstd(beat_upbeat_ratio(beats, upbeats, **kwargs))


def _validate_bur_inputs(beats: np.ndarray, upbeats: np.ndarray) -> None:
    """Validate beats and upbeats before calculating BURs"""
    if len(beats) < 2:
        raise ValueError("Must have at least two beats to calculate BURs")
    if len(upbeats) < 1:
        raise ValueError("Must have at least one upbeat to calculate BURs")
    if len(beats.shape) > 1 or len(upbeats.shape) > 1:
        raise ValueError("Beats and upbeats must be one-dimensional arrays")
    if np.isnan(np.sum(beats)) or np.isnan(np.sum(upbeats)):
        raise ValueError("Missing values found in beats or upbeats")
    for b1 in beats:
        if any(np.isclose(b1, upbeats, atol=TINY)):
            raise ValueError("Intersection found between beats and upbeats")


def match_beats_and_upbeats(
    beats: np.ndarray, upbeats: np.ndarray
) -> np.ndarray:
    """Iterates over consecutive beats and creates an array of `[[beat1, upbeat, beat2], [beat2, upbeat, beat3]]`"""

    matched = []
    # Iterate over consecutive pairs of beats
    for b1, b2 in zip(beats, beats[1:]):
        # Get the upbeats that are between both pairs of beats
        upbeat_bw = upbeats[(b1 < upbeats) & (upbeats < b2)]
        # Add a missing value in cases where multiple upbeats match with a single pair of beats, or no upbeats match
        #  We do not consider cases where multiple upbeats match with a single beat, as these are not "swing 8ths"
        #  Adding a missing value means that we'll have the shape len(beats) - 1
        if len(upbeat_bw) > 1 or len(upbeat_bw) == 0:
            matched.append([b1, None, b2])
        # This will only catch cases where we have matched a single upbeat
        else:
            matched.append([b1, upbeat_bw[0], b2])
    # Create a 2d array of [[beat1, upbeat, beat2], [beat2, upbeat, beat3]]
    return np.array(matched, dtype=float)


def _bur(beat1: float, upbeat: float, beat2: float) -> float:
    """BUR calculation function. Takes in two consecutive beats and gets BURs from provided upbeats."""

    # Everything should be in order
    assert beat1 - TINY < upbeat < beat2 + TINY
    # Calculate IOIs
    ioi1 = upbeat - beat1
    ioi2 = beat2 - upbeat
    # Calculate BUR
    return ioi1 / ioi2



"""pt_xml_import.py - Partitura XML import and conversion to AMADS"""

__author__ = "Roger B. Dannenberg"

import warnings
from math import isclose
from pathlib import Path
from typing import Optional

import partitura as pt
from partitura import score as ptScore

from amads.core.basics import (
    Chord,
    Clef,
    KeySignature,
    Measure,
    Note,
    Part,
    Rest,
    Score,
    Staff,
    TimeSignature,
)
from amads.core.pitch import Pitch

# plan: multiple passes over iter_all()
# for each part: add the part to a Concurrence
#      1st pass: get staff numbers from notes, extract measures, get div/qtr
#           create a Concurrence of staves if more than one,
#           create measures in each staff
#      2nd pass (A): get notes, rests, insert parameters into lists
#      2nd pass (B): build Note and Rest objects, insert into Measures
#      2nd pass (C): set ties attribute of Notes ?

# Partitura seems to have a rounding error, reporting measure length
# of 1919 instead of 1920 when divs per quarter is 480. This can lead
# to measures with fractional duration (e.g. 3.997916666666667) and
# the insertion of ties across measure boundaries that are in the
# wrong place. When DIV_TO_QUARTER_ROUNDING is not None, measure
# durations are rounded to the nearest 1/DIV_TO_QUARTER_ROUNDING of
# a quarter note. Do not round to whole beats because there might
# be a fractional measure with a pickup note. 24 allows for 32nd
# notes and 64th-note triplets (1/24 beat). Rounding must be enabled
# also by passing rnd=True to div_to_quarter(). The intent is to
# round measure boundaries but not note start/duration (because note
# times can be from a MIDI performance, where time is not quantized,
# at least not to symbolic durations or beats.)
#
DIV_TO_QUARTER_ROUNDING = 96

# algorithm: multiple passes over iter_all()
# for each part: add the part to a Concurrence
#      1st pass: get staff numbers from notes, extract measures, get div/qtr
#           create a Concurrence of staves if more than one,
#           create measures in each staff
#           build a map from every 10 beats to the Staff content index of a
#               measure that starts after that time. (to make it fast to find
#               measures - see description below)
#      2nd pass: get notes, rests, insert parameters into lists
#      3rd pass: re-tie notes that cross measures in case the measure boundary
#           has moved due to rounding
#      4th pass: build Note and Rest objects, insert into Measures
# measure_map: a list of indices into the Staff content list such that for any
#     time t, staff.content[measure_map[int(t/10)]].onset <= t and
#     staff.content[measure_map[int(t/10) + 1]].onset > t. Intuitively,
#     we find index i = int(t/10), and we're guaranteed that measure indexed by
#     measure_map[i] starts at or before time t. We can then search forward to
#     find the measure that contains t.
# Besides the measures map, we have a map pt_note_to_note from partitura note id
#     to both events and Note objects (as a list [event, note]) so we can find
#     the notes to adjust when we process ties.


def find_measure_ending_after(staff: Staff, time: float) -> Optional[Measure]:
    """Find the first measure in the staff that ends after time.
    staff - the Staff object
    time - the time to find the measure after

    Returns
    -------
      Measure - the first measure that ends after time
    """
    i = int(time / 10)
    if i >= len(measure_map):
        return None  # there are no measures to search
    i = measure_map[int(time / 10)]
    assert i < len(staff.content) and staff.content[i].onset <= time
    # now i is a known lower bound on the result we want
    i += 1
    while i < len(staff.content) and staff.content[i].onset <= time:
        i += 1
    # either staff.content[i] does not exist or onset of measure i is too high
    # either way, we want the previous measure:
    return staff.content[i - 1]


def div_to_quarter(durs, div, rnd=False):
    """given an array of (div, divs_per_qtr), map from div to quarter.
    This is not efficient if the array is large, i.e. if divs_per_qtr
    changes many times, which I assume is very unusual. Use this instead
    of quarter_map() because the latter maps partial first measure (pickup
    notes) to negative times, whereas we call the first event (rest or note)
    quarter 0. If rnd is True and DIV_TO_QUARTER_ROUNDING is not None,
    round the result to the nearest multiple of 1/DIV_TO_QUARTER_ROUNDING.
    """
    i = 0
    qtrs = 0
    while i + 1 < len(durs) and div > durs[i + 1][0]:
        # sum intervening quarters to this new time point:
        qtrs += (durs[i + 1][0] - durs[i][0]) / durs[i][1]
        i += 1
    # add quarters from last time point to "now" (div):
    qtrs += (div - durs[i][0]) / durs[i][1]
    if rnd and (DIV_TO_QUARTER_ROUNDING is not None):
        qtrs = round(qtrs * DIV_TO_QUARTER_ROUNDING) / DIV_TO_QUARTER_ROUNDING
    # print("div_to_quarter: div", div, "qtrs", qtrs)
    return qtrs


def if_tied(note: ptScore.Note) -> bool:
    """Given a partitura note, return the note itself if it is
    either tied-to or tied-from. Otherwise, return None.
    """
    return (note.tie_prev or note.tie_next) is not None


def retie_notes(event, staff, rnd: bool):
    """Adjust tied notes that cross barlines

    This method accounts for rounding in Partitura that can change
    measure timing. It adjusts the onsets and durations of tied notes
    so that ties occur at measure boundaries when they are already
    close, sometimes eliminating tied notes very small durations.

    event - the start of the tied note group
    staff - the staff containing the rest of the measures
    rnd - if True, apply rounding to measure boundaries (default True)

    Algorithm:
    If event is for a note tied to another note but not tied-from
    another note, e.g. it is the first in a tied group, find all tied
    notes using pt_note_to_note. Then for each note in order, if it is
    tied to the next note and the end time rounds to a bar time, then
    replace the end time with the bar time and replace the next note
    onset and duration to move the start time to the bar time. If the
    resulting duration is < 0.001, assume it was created through
    rounding error and set it to zero. (This could slightly truncate
    performance notes, but imperceptibly.) In the next pass, set the
    tied attribute to None if the tied-to note has duration 0 and ignore
    notes with duration of 0, which now indicates they have been deleted
    from a series of tied notes.
    """
    pt_note = event[7]
    if pt_note.tie_prev is not None or pt_note.tie_next is None:
        return

    # gather tied notes into group
    group = [event]
    while pt_note.tie_next is not None:
        pt_note = pt_note.tie_next
        # find the note in events
        ev = pt_note_to_note[pt_note][0]
        group.append(ev)

    for i, ev in enumerate(group[:-1]):  # check all but last event
        # does ev end near a measure boundary? If so assume it's a tie across
        # the bar:
        end = ev[1] + ev[2]

        # find the measure that ends after ev
        measure = find_measure_ending_after(staff, ev[1])
        assert measure is not None

        # see if the end of the note is near the end of the measure
        if rnd and abs(end - measure.offset) < 0.5 / DIV_TO_QUARTER_ROUNDING:
            # end of note rounds to the time of the end of measure
            extend = measure.offset - end  # could be >0 or <0
            ev[2] = measure.offset - ev[1]
            # group[i + 1] exists because we're iterating over group[:-1]
            group[i + 1][1] = measure.offset
            # if we extend ev==group[i], we need to shorten group[i+1]
            group[i + 1][2] -= extend
            # now if group[i+1] duration rounds to zero, we eliminate it
            if group[i + 1][2] < 0.5 / DIV_TO_QUARTER_ROUNDING:
                if group[i + 1][7].tie_next is not None:
                    warnings.warn(
                        "Unexpected very short note event in tied group "
                        + str(group[i + 1])
                    )
                group[i + 1][2] = 0  # indicate that note is removed
                break  # (maybe redundant, we should be done with iteration)


def staff_for_note(part: Part, event: list) -> Staff:
    """Find the staff corresponding to the Partitura staff number.
    part - the Part containing all staffs/staves.
    event - a tuple extracted from Partitura structure (see "Data formats..."
            below).
    Returns
    -------
      Staff - a Staff object from the Part that contains the event.
    """
    if event[3] is None:
        return part.content[0]  # type: ignore  (part has staff)
    else:
        return part.content[event[3] - 1]  # find the staff


def process_signatures(
    measure: Measure, staff_num: int, signatures: list[list]
) -> None:
    """
    If one or more signatures belong in measure, create them. This is
    called here because sometimes signatures appear before the Partitura
    measure, so we have to wait for the measure to be created before
    putting signature in the Score.

    Also, this functions sorts events into conventional order: clef < key_sig.

    The intent is to pass the list of signatures to every measure of every
    staff. For efficiency, the signatures are in order and as we process each
    measure, we remove any signature that we use or do not want. Since
    signatures is modified, a copy is made for each staff so that when we
    go to a new staff, we start with the full original list.

    Parameters
    ----------
    measure: Measure
        where to add the signature
    staff_num: int
        what staff does measure belong to?
    signatures: list
        list of tuples, currently either ("key_sig", onset, fifths) or
        ("clef", onset, staff_num, clef). "clef"s are included for all
        staffs, so this function skips

    """
    while len(signatures) > 0:
        sig = signatures[0]
        if sig[0] == "clef" and sig[2] != staff_num:
            del signatures[0]  # skip clefs for other staves
            continue
        if (
            len(signatures) > 1
            and signatures[1][0] == "clef"
            and signatures[1][2] != staff_num
        ):
            del signatures[1]  # skip clefs for other staves even when
            # there is a key signature at the front so that we can swap
            # clef and key signature if needed below
        if sig[1] >= measure.onset:
            # "cheap and dirty" sort since there are only 2 categories:
            if (
                len(signatures) > 1
                and isclose(sig[1], signatures[1][1], abs_tol=0.001)
                and sig[0] == "key_sig"
                and signatures[1][0] == "clef"
                and signatures[1][2] == staff_num
            ):
                signatures[0] = signatures[1]  # swap 0 and 1
                signatures[1] = sig
                sig = signatures[0]
            if sig[0] == "key_sig":
                KeySignature(measure, sig[1], sig[2])
            elif sig[0] == "clef" and sig[2] == staff_num:
                Clef(measure, sig[1], sig[3])
            else:
                assert False, f"Internal error, unexpected sig {sig}"
            del signatures[0]
        else:
            return  # need to wait for measure to be created


def partitura_convert_part(
    ppart: ptScore.Part, score: Score, rnd: bool = True
) -> Part:
    """Convert a Partitura part to an AMADS Part and add it to the Score.
    ppart - the Partitura part
    score - the AMADS Score to which the Part will be added
    rnd - if True, does some select rounding for symbolic scores.
          if False (default), no rounding is done (used for MIDI import)
    Returns
    -------
      Part - the newly created AMADS Part
    """
    # these are globals so we don't have to pass to every
    # helper function that needs to do lookups:
    global measure_map, pt_note_to_note
    # note ppart.quarter_map(x) maps divs to quarters
    part: Part = Part(parent=score, instrument=ppart.part_name)
    durs = ppart.quarter_durations()
    staff_numbers = set()
    # data is stored in ppart in a different order than we want, so
    # we first extract it into various lists, each of which will be
    # accessed in order. (This also means only one iteration of ppart
    # because iter_all() is currently extremely slow, so we only want
    # to do it once.) We traverse measures and signatures for each
    # staff in a subsequent pass.
    measures = []  # list of (measure number, start time, end time) tuples
    signatures = []  # list of ("key_sig", onset, fifths) and
    # ("clef", onset, staff_num, clef) info
    notes = []  # list of ("note", ...), ("rest", ...) or ("tempo", ...)
    # information
    measure_map = []
    pt_note_to_note = {}

    # pass 1: count staves and collect measure, signature, notes lists
    tied_count = 0  # debug to count notes in part
    for item in ppart.iter_all():
        if isinstance(item, ptScore.Note) or isinstance(item, ptScore.Rest):
            staff_numbers.add(item.staff)

        onset = div_to_quarter(durs, item.start.t)  # type: ignore
        if isinstance(item, ptScore.Measure):
            # convert divs duration to quarters
            duration = div_to_quarter(
                durs, item.end.t, rnd=rnd  # type: ignore
            ) - div_to_quarter(
                durs, item.start.t, rnd=rnd  # type: ignore
            )  # type: ignore
            if duration > 0:  # all staves have the same
                # measure count and timing, so we only build the map for
                # staff 0; do not append zero-length measures that arise
                # from rounding errors in Partitura:
                #
                # When a measure onset will land on or exceed a 10-beat
                # boundary, add a map entry. Use while in case measures are
                # longer than 10 beats, which means we'll have multiple entries
                # denoting to the same measure.
                #
                # we want staff.content[measure_map[int(t/10)]].onset <= t,
                # i.e. measure_map[0].onset == 0, measure_map[1].onset <= 10,
                # measure_map[2].onset <= 20, etc.
                offset = onset + duration
                while offset >= len(measure_map) * 10:
                    measure_map.append(len(measures))
                measures.append((onset, duration))
        elif isinstance(item, ptScore.TimeSignature):
            upper = item.beats
            lower = item.beat_type
            # find the time signature in effect at onset, add small offset
            # to avoid floating point rounding errors:
            ts = score._find_time_signature(onset + 1e-6)
            if ts.upper != upper or ts.lower != lower:
                last_ts = score.time_signatures[-1]
                if last_ts.time > onset:
                    warnings.warn(
                        "Encountered a new Partitura time signature"
                        " placed BEFORE an earlier time signature:"
                        " {element}. Something is probably wrong"
                        " with this score. Ignoring the time"
                        " signature in conversion to AMADS."
                    )
                else:
                    ts = TimeSignature(onset, upper, lower)
                    score.append_time_signature(ts)
        elif isinstance(item, ptScore.KeySignature):
            signatures.append(("key_sig", onset, item.fifths))
        elif isinstance(item, ptScore.Clef):
            clef = None
            if not item.octave_change or item.octave_change == 0:
                if item.sign == "G":
                    if item.line == 2:
                        clef = "treble"
                elif item.sign == "F":
                    if item.line == 4:
                        clef = "bass"
                elif item.sign == "C":
                    if item.line == 3:
                        clef = "alto"
                    elif item.line == 4:
                        clef = "tenor"
            elif (
                item.octave_change == -1 and item.sign == "G" and item.line == 2
            ):
                clef = "treble8vb"
            elif item.sign == "percussion":
                clef = "percussion"
            if clef is None:
                warnings.warn(
                    f'Unrecognized Partitura clef "{item}"'
                    f" octave_change {item.octave_change} ignored."
                )
            else:
                signatures.append(("clef", onset, item.staff, clef))
        elif isinstance(item, ptScore.Note):
            qtr = item.start.quarter  # type: ignore
            duration = (item.end.t - item.start.t) / qtr  # type: ignore
            is_tied = if_tied(item)
            notes.append(
                [
                    "note",
                    onset,
                    duration,
                    item.staff,
                    item.midi_pitch,
                    item.id,
                    is_tied,  # event[6]
                    item,
                ]
            )  # event[7]
            if is_tied:
                pt_note_to_note[item] = [notes[-1]]
                tied_count += 1  # debug
        elif isinstance(item, ptScore.Rest):
            qtr = item.start.quarter  # type: ignore
            duration = (item.end.t - item.start.t) / qtr  # type: ignore
            notes.append(["rest", onset, duration, item.staff])
        elif isinstance(item, ptScore.Tempo):
            # Note: partitura "bpm" is really beats per second!
            factor = 1.0  # conversion factor from partitura bpm to AMADS qps
            if item.unit is not None:  # unit string provided
                if item.unit == "h":
                    factor = 2.0  # half notes per second
                elif item.unit == "q":
                    factor = 1.0  # quarter notes per second
                elif item.unit == "q.":
                    factor = 1.5  # dotted quarter notes per second
                elif item.unit == "e":
                    factor = 0.5  # eighth notes per second
                elif item.unit == "e.":
                    factor = 0.75  # dotted eighth notes per second
                elif item.unit == "s":
                    factor = 4.0  # sixteenth notes per second
                elif item.unit == "s.":
                    factor = 0.375  # dotted sixteenth notes per second
                else:
                    ValueError(f"Unknown tempo unit in partitura: {item.unit}")
            score.time_map.append_change(onset, item.bpm * factor)

    # for each staff, create measures
    for staff_num in staff_numbers:
        # not sure what staff_num can be, so trying to be robust here:
        if staff_num is not None and not isinstance(staff_num, int):
            staff_num = str(staff_num)
            assert str(staff_num).isdigit(), (
                "Unexpected staff value" " from Partitura note"
            )
            staff_num = int(staff_num)
        staff = Staff(parent=part, number=staff_num)
        # staff_signatures will be "consumed" by new measures, so make a copy:
        staff_signatures = signatures.copy()
        for m_info in measures:
            m = Measure(parent=staff, onset=m_info[0], duration=m_info[1])
            process_signatures(m, staff_num, staff_signatures)
        staff.inherit_duration()

    # Data formats (note that staff number is always event[3])
    #    (Note, onset, duration, staff, midi_pitch, id, tied, ptnote)
    #    (Rest, onset, duration, staff)
    # id_to_event = {}

    # pass 3: re-tie notes that cross measures in case measure times
    #         have changed due to rounding.
    if rnd:
        for i, event in enumerate(notes):
            staff = staff_for_note(part, event)
            if event[0] == "note" and event[6]:  # tied
                retie_notes(event, staff, rnd)

    # pass 4: insert notes and rests into score
    mindex = 0
    for i, event in enumerate(notes):
        staff = staff_for_note(part, event)
        measure: Measure = staff.content[mindex]  # type: ignore
        # find the measure containing event[1] (onset)
        while event[1] >= measure.offset:
            mindex += 1
            if mindex == len(staff.content):
                warnings.warn(
                    f"Something is wrong; could not find measure"
                    f" for {event}"
                )
                break  # use previous measure, but probably there is a bug here
            measure = staff.content[mindex]  # type: ignore
        if event[0] == "note":
            if event[2] > 0:  # zero duration means skip note
                note = Note(
                    parent=measure,
                    onset=event[1],
                    duration=event[2],
                    pitch=Pitch(event[4]),
                )
                if event[6]:  # is tied to another note
                    # Multiple cases: 1) note is tied to next note with
                    # non-zero duration, so we put the note in pt_note_to_note
                    # so it can be patched later. 2) note is tied to a previous
                    # note, so we patch the previous note.
                    pt_note = event[7]
                    # map pt_note to [event, note], so [0] gives the event,
                    # and event[2] is duration
                    if pt_note.tie_next and pt_note_to_note[pt_note][0][2] != 0:
                        # associate this new Note with the partitura note:
                        pt_note_to_note[pt_note].append(note)
                    if pt_note.tie_prev:
                        # patch the previous note
                        pt_note = pt_note.tie_prev
                        pt_note_to_note[pt_note][1].tie = note
        elif event[0] == "rest":
            Rest(parent=measure, onset=event[1], duration=event[2])
        else:
            assert False, f"Unknown event type {event[0]}"

    # expand first measures to a full measure if necessary
    # what is the maximum offset of the first measure?
    for staff in part.content:
        if len(staff.content) > 0:  # type: ignore
            m1 = staff.content[0]  # type: ignore
            max_offset = 0
            for elem in m1.content:
                max_offset = max(max_offset, elem.offset)
            m1_ts_dur = m1.time_signature().quarters
            if max_offset < m1_ts_dur - 0.001:  # need to insert rest
                gap = m1_ts_dur - max_offset
                for elem in m1.content:
                    if (
                        isinstance(elem, Note)
                        or isinstance(elem, Rest)
                        or isinstance(elem, Chord)
                    ):
                        elem.time_shift(gap)
                # insert Rest, Since m1 is first, m1.onset == 0
                _ = Rest(m1, m1.onset, duration=gap)

                # if m1 duration increases, shift all other measures
                gap = m1_ts_dur - m1.offset
                if gap > 0.001:
                    m1.duration = m1_ts_dur  # fix m1 duration
                    staff.pack()  # type: ignore

                # also need to update time map and time_signatures
                for ts in score.time_signatures[1:]:
                    ts.time += gap
                # we're adding gap beats at the initial tempo,
                # so convert that to a time shift in seconds,
                # then add (time_gap, gap) to (time, quarter) pairs
                time_gap = gap * score.time_map.get_tempo_at(0)
                for mq in score.time_map.changes[1:]:
                    mq.time += time_gap
                    mq.quarter += gap

    # last measure of Partitura can be "incomplete" and have a duration
    # shorter than what is implied by the time signature. Here, we will
    # "expand" the last measures to full time signature duration, consistent
    # with expanding incomplete first measures and with Music21 import.
    # We assume every part has the same structure so we can adjust them
    # independently:
    offset = 0.0
    for staff in part.content:
        if len(staff.content) > 0:  # type: ignore
            mn = staff.content[-1]  # type: ignore  last measure
            mn_ts_dur = mn.time_signature().quarters
            if mn.duration < mn_ts_dur - 0.001:
                mn.duration = mn_ts_dur
                offset = max(offset, mn.offset)
    for staff in part.content:
        if offset > staff.offset:
            staff.offset = offset

    part.duration = part.content[0].duration  # set to equal staff duration
    return part


def partitura_import(
    filename: Path | str,
    format: str,
    flatten: bool = False,
    collapse: bool = False,
    show: bool = False,
    group_by_instrument: bool = True,
) -> Score:
    """Use Partitura to import a MusicXML file.

    Parameters
    ----------
    filename : Path | str
        The path (relative or absolute) to the music file.
    format: str
        The format of the file: 'musicxml', 'mei', 'kern'
    flatten : bool = False
        Returns a flat score (see `amads.core.basics.Part.flatten`)
    collapse : bool = False
        If `flatten` is True, all Notes are merged (collapsed) into
        a single Part in the resulting Score.
    show : bool = False
        Print a text representation of the data.
    group_by_instrument : bool = True
        This parameter is ignored by Partitura, which automatically
        produces parts with multiple staffs. The Partitura grouping
        is respected, and `group_by_instrument` is ignored.

    Returns
    -------
    Score
        The imported Score

    <small>**Author**: Roger B. Dannenberg</small>
    """
    # Partitura.load_score claims to accept a file-like object, but
    # it has been problematic for us in the past, so we use a string filename.
    filename = str(filename)

    # maybe load_score would handle all file types, but since we know the
    # format and might want to use some format-dependent parameters, handle
    # each case separately:
    ptscore = None
    if format == "musicxml":
        ptscore = pt.load_musicxml(filename)
    elif format == "mei":
        ptscore = pt.load_mei(filename)
    elif format == "kern":
        ptscore = pt.load_kern(filename)
    assert ptscore is not None, (
        "Partitura failed to load score from " f"{filename}"
    )
    if show:
        print(f"Partitura score structure from {filename}:")
        for ptpart in ptscore:
            print(ptpart.pretty())
    score = Score()
    for ptpart in ptscore.parts:
        partitura_convert_part(ptpart, score)
    score.inherit_duration()
    if flatten or collapse:
        score = score.flatten(collapse=collapse)
    return score



# -*- coding: utf-8 -*-
"""
Parsons code for contour of musical melody by direction only.
"""

from typing import Optional

from amads.core.utils import sign

__author__ = "Mark Gotham"


class ParsonsContour:
    """Implementation of the basic Parsons contour classification scheme.

    Parsons categorises each step by direction only.

    Nothing more, nothing less.

    <small>**Author**: Mark Gotham</small>
    """

    def __init__(
        self,
        pitches: list[int],
        character_dict: Optional[dict] = None,
        initial_asterisk: bool = False,
    ):
        """
        The 'Parsons code' returns simply the direction of each successive melodic interval.
        It's been used in lookup, and can serve as a useful first entry point to the topic of contour.

        Parameters
        ----------
        pitches:
            A list of integers representing pitches
            (assumed to be MIDI numbers or equivalent, not pitch classes)
        character_dict:
            A dict specifying which characters to use when mapped to a string.
            Must include keys for [1, 0, -1] corresponding to up, repeat, and down.
            The default is Parsons' own values: {1: "u", 0: "r", -1: "d"}.
            Other options could include `<`, `=`, `>`.
        initial_asterisk:
            Optionally, include an initial `*` for the start of the sequence (no previous interval).

        Examples
        --------
        >>> happy = [60, 60, 62, 60, 65, 64, 60, 60, 62, 60, 67, 65, 60, 60, 72, 69, 65, 64, 62, 70, 69, 65, 67, 65]
        >>> pc = ParsonsContour(happy)
        >>> pc.interval_sequence
        [None, 0, 2, -2, 5, -1, -4, 0, 2, -2, 7, -2, -5, 0, 12, -3, -4, -1, -2, 8, -1, -4, 2, -2]

        >>> pc.interval_sequence_sign
        [None, 0, 1, -1, 1, -1, -1, 0, 1, -1, 1, -1, -1, 0, 1, -1, -1, -1, -1, 1, -1, -1, 1, -1]

        >>> pc.as_string
        'rududdrududdrudddduddud'

        >>> twinkle_ints = [72, 72, 79, 79, 81, 81, 79, 77, 77, 76, 76, 74, 74, 72]
        >>> pc = ParsonsContour(twinkle_ints)
        >>> pc.as_string
        'rururddrdrdrd'

        >>> pc_no_asterisk = ParsonsContour(twinkle_ints, initial_asterisk=True)
        >>> pc_no_asterisk.as_string
        '*rururddrdrdrd'

        >>> pc_symbols = ParsonsContour(twinkle_ints, {1: "<", 0: "=", -1: ">"})
        >>> pc_symbols.as_string
        '=<=<=>>=>=>=>'

        References
        ----------
        [1] Parsons, Denys. 1975. *The Directory of Tunes and Musical Themes*.
        """

        self.pitches = pitches
        self.character_dict = (
            character_dict if character_dict else {1: "u", 0: "r", -1: "d"}
        )
        self.initial_asterisk = initial_asterisk

        self.interval_sequence = None
        self.interval_sequence_sign = None
        self.as_string = None

        self.get_intervals()
        self.make_string()

    def get_intervals(self):
        self.interval_sequence = [
            None
        ]  # Now as lists, still initialised with None, as per AMADS policy ...
        self.interval_sequence_sign = [None]
        for i in range(len(self.pitches) - 1):
            this_diff = self.pitches[i + 1] - self.pitches[i]
            self.interval_sequence.append(this_diff)
            self.interval_sequence_sign.append(sign(this_diff))

    def make_string(self):
        """Create a flat, string representation of the contour directions."""
        self.as_string = ""
        if self.initial_asterisk:
            self.as_string += "*"
        for i in range(1, len(self.interval_sequence_sign)):
            self.as_string += self.character_dict[
                self.interval_sequence_sign[i]
            ]


# ------------------------------------------------------------------------------

if __name__ == "__main__":
    import doctest

    doctest.testmod()



"""
What does it mean for music to be "fast" or "slow"?
Certainly BPM is not enough.
The "Attractor tempos" theory (Gotham 2015, [1]) proposes
 a definition of "fast"/"slow" relative to neutral, central, moderate tempos
 and a definition of those moderate ("Attractor") tempos that accounts for the role of the metrical structure.
In short, it provides a model for optimizing the salience of metrical structures.

<small>**Author**: Mark Gotham</small>

Reference
---------
[1] Gotham, M. (2015). Attractor tempos for metrical structures. Journal of Mathematics and Music, 9(1), 23–44.
https://doi.org/10.1080/17459737.2014.980343
"""

__author__ = "Mark Gotham"


from typing import Optional

import matplotlib.pyplot as plt
import numpy as np

# ------------------------------------------------------------------------------


class MetricalSalience:
    """
    Methods for storing array representations of metrical structure and derived salience values.

    Parameters
    ----------
    symbolic_pulses:
        A NumPy array representing the symbolic pulse lengths by level.
    quarter_bpm:
        The beats-per-minute corresponding to the symbolic value of a pulse length 1.0 in symbolic time.
        The user sets this value if/when calculating absolute length and salience values.
    mu: float
        The mean of the Gaussian.
    sig: float
        The standard deviation of the Gaussian.

    Attributes
    ----------
    symbolic_pulses:
        As above.
    absolute_pulses:
        An adaptation of the symbolic pulse lengths array that maps each value from symbolic to seconds.
    salience_values:
        An adaptation of the absolute pulse lengths to the equivalent salience values (see notes on `log_gaussian`).
    cumulative_salience_values:
        A 1D array summation of the absolute salience values by column (one value per metrical position).
    indicator:
        An indicator array for the (non-)presence of values at each position of the symbolic pulse lengths array.
        This can serve, for example, as the symbolic equivalent of the (absolute) `salience_values` array.

    Examples
    --------

    >>> from amads.time.meter.representations import PulseLengths
    >>> pl = [4, 2, 1, 0.5]
    >>> pls = PulseLengths(pulse_lengths=pl, cycle_length=4)
    >>> arr = pls.to_array()
    >>> arr
    array([[4. , 0. , 0. , 0. , 0. , 0. , 0. , 0. ],
           [2. , 0. , 0. , 0. , 2. , 0. , 0. , 0. ],
           [1. , 0. , 1. , 0. , 1. , 0. , 1. , 0. ],
           [0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5]])

    >>> ms = MetricalSalience(symbolic_pulses=arr, quarter_bpm=120)
    >>> ms.absolute_pulses
    array([[2.  , 0.  , 0.  , 0.  , 0.  , 0.  , 0.  , 0.  ],
           [1.  , 0.  , 0.  , 0.  , 1.  , 0.  , 0.  , 0.  ],
           [0.5 , 0.  , 0.5 , 0.  , 0.5 , 0.  , 0.5 , 0.  ],
           [0.25, 0.25, 0.25, 0.25, 0.25, 0.25, 0.25, 0.25]])

    >>> ms.cumulative_salience_values
    array([2.39342011, 0.44793176, 1.4136999 , 0.44793176, 2.17446773,
           0.44793176, 1.4136999 , 0.44793176])

    """

    def __init__(
        self,
        symbolic_pulses: Optional[np.ndarray] = None,
        quarter_bpm: Optional[float] = None,
        mu: float = 0.6,
        sig: float = 0.3,
    ):
        self.symbolic_pulses = symbolic_pulses
        self.quarter_bpm = quarter_bpm
        self.mu = mu
        self.sig = sig
        self.absolute_pulses = self.calculate_absolute_pulse_lengths()
        self.salience_values = self.calculate_salience_values()
        self.cumulative_salience_values = (
            self.calculate_cumulative_salience_values()
        )
        self.indicator = self.make_indicator()

    def calculate_absolute_pulse_lengths(self):
        """
        Calculate absolute pulse lengths from
        the symbolic lengths (`symbolic_pulses`) and
        the BPM provided here for the 'quarter note' as reference value.
        """
        return self.symbolic_pulses * (60 / self.quarter_bpm)

    def calculate_salience_values(self):
        """
        Calculate salience values for items in the `symbolic_pulses`
        using `log_gaussian` (see notes on that function).
        """
        return log_gaussian(self.absolute_pulses, self.mu, self.sig)

    def calculate_cumulative_salience_values(self):
        """
        Calculate cumulative salience values by summing over columns.
        """
        return np.sum(self.salience_values, axis=0)

    def make_indicator(self):
        """
        Make a 2D indicator vector for the presence/absense of a pulse value at each position.
        """
        return (self.symbolic_pulses > 0).astype(int)

    def plot(
        self,
        symbolic_not_absolute: bool = False,
        reverse_to_plot: bool = True,
        show: bool = True,
    ):
        """
        Plot the salience values with their respective contribution.

        Parameters
        ----------
        symbolic_not_absolute: If True, plot only the indicator values (one per level).
            If False (default), plot the tempo- and meter-sensitive, weighted salience values.
        reverse_to_plot: If True (default), plot the fastest values at the bottom.

        Returns
        -------
        Figure
            A matplotlib.figure.Figure of the plotted salience values.
        """
        if symbolic_not_absolute:
            data = self.indicator
        else:
            data = self.salience_values

        pulse_values_for_labels = self.symbolic_pulses[:, 0]

        if reverse_to_plot:
            data = data[::-1]  # TODO maybe revisit for elegance, checks
            pulse_values_for_labels = pulse_values_for_labels[::-1]

        num_layers = data.shape[0]
        num_cols = data.shape[1]
        fig, ax = plt.subplots()
        bottom = np.zeros(num_cols)

        for i in range(num_layers):
            ax.bar(
                np.arange(num_cols),
                data[i],
                bottom=bottom,
                label=f"Pulse={pulse_values_for_labels[i]}; IOI={pulse_values_for_labels[i] * 60 / self.quarter_bpm}",
            )
            bottom += data[i]

        ax.set_xlabel("Cycle-relative position")
        ax.set_ylabel("Weighting")
        ax.legend()
        ax.grid(True)
        if show:
            plt.show()
        return fig


def log_gaussian(arr: np.ndarray, mu: float = 0.6, sig: float = 0.3):
    """
    Compute a log-linear Gaussian which is the basis of individual pulse salience values.
    To avoid log(0) issues, `np.clip` values to be always greater than 0.
    See also [MetricalSalience.calculate_salience_values]
    [amads.time.meter.attractor_tempos.MetricalSalience.calculate_salience_values].


    Parameters
    ----------
    mu: float
        The mean of the Gaussian.
    sig: float
        The standard deviation of the Gaussian.

    Examples
    --------

    >>> log_gaussian(np.array([0.06, 0.6, 6.0])) # demo log-lin symmetry
    array([0.00386592, 1.        , 0.00386592])

    >>> log_gaussian(np.array([0.5, 1., 2.])) # 2x between levels
    array([0.96576814, 0.76076784, 0.21895238])

    """
    if sig <= 0:
        raise ValueError("Standard deviation (`sig`) must be positive.")
    if mu <= 0:
        raise ValueError("Mean (`mu`) must be positive.")
    x = np.clip(arr, 1e-9, None)
    return np.exp(-(np.log10(x / mu) ** 2 / (2 * sig**2)))


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from math import isclose
from pathlib import Path
from typing import Optional, cast

from partitura import save_musicxml
from partitura.score import Clef as ptClef
from partitura.score import GenericNote as ptGenericNote
from partitura.score import KeySignature as ptKeySignature
from partitura.score import Measure as ptMeasure
from partitura.score import Note as ptNote
from partitura.score import Part as ptPart
from partitura.score import Rest as ptRest
from partitura.score import Score as ptScore
from partitura.score import Staff as ptStaff
from partitura.score import TimeSignature as ptTimeSignature

from amads.core.basics import (
    Clef,
    EventGroup,
    KeySignature,
    Measure,
    Note,
    Part,
    Rest,
    Score,
    Staff,
)

DIVS = 600  # divisions per quarter note for Partitura MIDI export
# 600 corresponds to 1 ms at 100 bpm

CLEF_SIGN = {
    "treble": "G",
    "bass": "F",
    "alto": "C",
    "tenor": "C",
    "percussion": "percussion",
    "treble8vb": "G",
}
CLEF_LINE = {
    "treble": 2,
    "bass": 4,
    "alto": 3,
    "tenor": 4,
    "percussion": None,
    "treble8vb": 2,
}
CLEF_OCTAVE_CHANGE = {
    "treble": None,
    "bass": None,
    "alto": None,
    "tenor": None,
    "percussion": None,
    "treble8vb": -1,
}


def is_new_key_signature(event: KeySignature, kstimes: list[float]) -> bool:
    for kstime in kstimes:
        if isclose(kstime, event.onset, abs_tol=0.001):
            return False
    kstimes.append(event.onset)
    return True


def add_event_to_part(
    event,
    pt_part: ptPart,
    kstimes: list[float],
    staff: Optional[int],
    id: int,
    ties: Optional[dict] = None,
) -> int:
    """Add an AMADS event to a Partitura part.

    This is a recursive function that handles different event types.
    Partitura requires Staff objects to have IDs and Notes to have
    Staff ID's (if there are multiple staves), so we pass in the
    staff ID when iterating through a Staff's content, and to
    support assigning unique IDs to new Staffs, we pass in a
    "suggested" object ID; if it is used, we return the ID + 1
    to be passed to the next Staff in the (Part's) content.

    Also, Partitura numbers measures starting at 1, but AMADS
    allows measures without numbers and allows strings as measure
    numbers; we assign our own count to measures and assign
    AMADS measure numbers to the Partitura Measure name attribute.

    Partitura does not support per-staff key signatures, so kstimes
    is used to remember when key signatures have already been
    specified to eliminate adding duplicates.

    ties: dict mapping from AMADS Note to (partitura Note, is_tied_to)
    where is_tied_to indicates if there's an incoming tie from a previous note.
    """
    pt_group = None
    if isinstance(event, Staff):  # must be full score
        pt_group = ptStaff(number=id)
        staff = id
        id += 1  # so we can return an id for the next Staff
    # add measures, but only if no staff or in first staff:
    elif isinstance(event, Measure) and (staff is None or staff == 1):
        event = cast(Measure, event)
        name = None if event.number is None else str(event.number)
        pt_group = ptMeasure(number=id, name=name)  # no staff here
    if isinstance(event, EventGroup):  # e.g. Part, Staff, Measure
        if pt_group is not None:
            pt_part.add(
                pt_group, round(event.onset) * DIVS, round(event.offset) * DIVS
            )
        subid = 1
        for subevent in event.content:
            subid = add_event_to_part(
                subevent, pt_part, kstimes, staff, subid, ties
            )
    elif isinstance(event, Note):
        event = cast(Note, event)
        if event.pitch is None:
            pt_note = ptGenericNote(staff=staff, voice=staff)
        else:
            pt_note = ptNote(
                step=event.step,
                octave=event.octave,
                alter=event.pitch.alt,
                staff=staff,
                voice=staff,
            )  # type: ignore
            assert pt_note.midi_pitch == event.key_num, (
                "internal error in pitch"
                " conversion; maybe octave confusion for something like B#3?"
            )
        pt_part.add(
            pt_note, round(event.onset * DIVS), round(event.offset * DIVS)
        )
        # Track tied notes
        if ties is not None:
            if event.tie:  # This note is tied to event.tie
                is_tied_to = (event in ties) and ties[event][1]
                ties[event] = (pt_note, is_tied_to)
                # Mark the note this ties to as having an incoming tie
                if event.tie in ties:
                    ties[event.tie] = (ties[event.tie][0], True)
                else:
                    ties[event.tie] = (None, True)
            elif event in ties:
                # Update entry with the pt_note we just created
                ties[event] = (pt_note, ties[event][1])
    elif isinstance(event, Rest):
        event = cast(Rest, event)
        pt_rest = ptRest(staff=staff)
        pt_part.add(
            pt_rest, round(event.onset * DIVS), round(event.offset * DIVS)
        )
    elif isinstance(event, KeySignature) and is_new_key_signature(
        event, kstimes
    ):
        event = cast(KeySignature, event)
        pt_key_sig = ptKeySignature(fifths=event.key_sig, mode=None)  # no staff
        pt_part.add(pt_key_sig, round(event.onset * DIVS))
    elif isinstance(event, Clef):
        event = cast(Clef, event)
        pt_clef = ptClef(
            staff=staff,
            sign=CLEF_SIGN[event.clef],
            line=CLEF_LINE[event.clef],
            octave_change=CLEF_OCTAVE_CHANGE[event.clef],
        )
        pt_part.add(pt_clef, round(event.onset * DIVS))
    return id


def score_to_partitura(
    score: Score,
    filename: Path | str,
    show: bool = False,
) -> ptScore:
    """Convert an AMADS Score to a Partitura score."""
    parts = []
    previous_part_number = 0
    ties = {}  # Map from AMADS Note to (partitura Note, is_tied_to)
    # where partitura Note corresponds to the key (AMADS Note)
    for part in score.find_all(Part):
        part = cast(Part, part)
        part_number = part.number
        if part_number is None:
            previous_part_number += 1
            part_number = previous_part_number
        else:
            part_number = int(part_number)
            previous_part_number = max(previous_part_number, part_number)
        # Partitura expects part IDs as strings (used as MusicXML id attributes)
        pt_part = ptPart(
            id=str(part_number),
            part_name=part.instrument,
            quarter_duration=DIVS,
        )
        parts.append(pt_part)

        # add time signatures
        for ts in score.time_signatures:
            # partitura does not allow float/fractional beats/measure
            pt_ts = ptTimeSignature(round(ts.upper), ts.lower)
            pt_part.add(pt_ts, round(ts.time * DIVS))  # no end time?

        # add content
        id = 1  # At this level, if event is a Staff, id becomes staff number
        kstimes = []  # key signature onset times
        for event in part.content:
            id = add_event_to_part(
                event, pt_part, kstimes, staff=None, id=id, ties=ties
            )

    pt_score = ptScore(partlist=parts)

    # Fix up ties: set tie_next and tie_prev on partitura notes
    for amads_note in ties.keys():
        (pt_note, is_tied_to) = ties[amads_note]

        if pt_note is None:
            # Note was tied to but we never created its partitura note
            continue

        # Set tie_next if this AMADS note is tied to another
        if amads_note.tie and amads_note.tie in ties:
            tied_to_pt_note = ties[amads_note.tie][0]
            if tied_to_pt_note is not None:
                pt_note.tie_next = tied_to_pt_note
                tied_to_pt_note.tie_prev = pt_note

    if show:
        print(f"Partitura score structure for {str(filename)}.")
        print(f"---- class is {pt_score.__class__}, here are the parts ----")
        for ptpart in pt_score.parts:
            print(ptpart.__class__)
            print(ptpart.pretty())
    return pt_score


def partitura_export(
    score: Score,
    filename: Path | str,
    format: str,
    show: bool = False,
    display: bool = False,
) -> None:
    """Save a Score to a file in musicxml format using Partitura.

    <small>**Author**: Roger B. Dannenberg</small>

    Parameters
    ----------
    score : Score
        The Score to export.
    filename : Path | str
        The name or path of the file to save the MusicXML data.
    format : str
        The format of the file to save. Must be "musicxml" for this function.
    show : bool, optional
        If True, print the partitura score structure for debugging.
    display : bool, optional
        If True, open the generated PDF in the default viewer.
        Only relevant for PDF export (see `music21_xml_lilypond_export`).
    """
    ptscore = score_to_partitura(score, filename, show)
    assert (
        format == "musicxml"
    ), f"Unsupported format for partitura export: {format}"
    save_musicxml(ptscore, filename)



"""
Import all public members from all amads submodules.

This module provides explicit imports from all amads submodules,
making all their public exports available at the package level.
"""

# algorithms
from .algorithms.complexity import *
from .algorithms.entropy import *
from .algorithms.gcd import *
from .algorithms.mtype_tokenizer import *
from .algorithms.ngrams import *
from .algorithms.nnotes import *
from .algorithms.norm import *
from .algorithms.scale import *
from .algorithms.slice.salami import *
from .algorithms.slice.slice import *
from .algorithms.slice.window import *

# core
from .core.basics import *
from .core.distribution import *
from .core.histogram import *
from .core.pitch import *
from .core.timemap import *
from .core.utils import *
from .core.vector_transforms_checks import *
from .core.vectors_sets import *

# harmony
from .harmony.consonance.consonance import *
from .harmony.root_finding.parncutt import *
from .io.pianoroll import *
from .io.readscore import *

# io
from .io.writescore import *

# melody
from .melody.boundary import *
from .melody.contour.huron_contour import *
from .melody.contour.interpolation_contour import *
from .melody.contour.parsons_contour import *
from .melody.contour.polynomial_contour import *
from .melody.contour.step_contour import *
from .melody.fantastic import *
from .melody.segment import *
from .melody.segment_gestalt import *
from .melody.similarity.melsim import *

# music
from .music import example
from .music.example import *

# pitch
from .pitch.hz2midi import *
from .pitch.ismonophonic import *
from .pitch.ivdirdist1 import *
from .pitch.ivdist1 import *
from .pitch.ivdist2 import *
from .pitch.ivsizedist1 import *
from .pitch.key.key_cc import *
from .pitch.key.keymode import *
from .pitch.key.keysom import *
from .pitch.key.keysomdata import *
from .pitch.key.kkcc import *
from .pitch.key.kkkey import *
from .pitch.key.max_key_cc import *
from .pitch.key.profiles import *
from .pitch.key.transpose2c import *
from .pitch.pc_set_functions import *
from .pitch.pc_sets import *
from .pitch.pcdist1 import *
from .pitch.pcdist2 import *
from .pitch.pitch_mean import *
from .pitch.serial import *
from .pitch.transformations import *

# polyphony
from .polyphony.skyline import *

# schema
from .schema.partimenti import *

# time
from .time.durdist1 import *
from .time.durdist2 import *
from .time.meter.attractor_tempos import *
from .time.meter.break_it_up import *
from .time.meter.examples import *
from .time.meter.grid import *
from .time.meter.profiles import *
from .time.meter.representations import *
from .time.meter.syncopation import *
from .time.meter.tatum import *
from .time.notedensity import *
from .time.rhythm import *
from .time.swing import *
from .time.tempo import *
from .time.variability import *



"""
Basic, local calculations of the greatest common divisor (GCD)
at a central point in the code base for wide applicability.
Parameter-specific adaptations are in the relevant place (e.g., `amads.time.meter`).

The module is organised in a kind of logical quadrant for
data type (e.g., integer vs fraction)
and
number (a single pair vs many pairwise calculations).

<small>**Author**: Mark Gotham</small>
"""

__author__ = "Mark Gotham"


from collections.abc import Sequence
from fractions import Fraction

# -----------------------------------------------------------------------------

# Direct operations on pair of values, known to be integer, float, fraction.


def integer_gcd_pair(a: int, b: int) -> int:
    """
    Calculates the greatest common divisor (GCD) of two integers using the
    Euclidean algorithm.

    >>> integer_gcd_pair(0, 2)
    2

    >>> integer_gcd_pair(15, 16)
    1

    >>> integer_gcd_pair(8, 16)
    8

    """
    while b:
        a, b = b, a % b
    return a


def float_gcd_pair(a: float, b: float = 1.0, rtol=1e-05, atol=1e-08) -> float:
    """
    Calculate approximate greatest common divisor (GCD) for values a and b given the specified
    relative and absolute tolerance (`rtol` and `atol`).
    With thanks to Euclid,
    `fractions.gcd`, and
    `stackexchange <https://stackoverflow.com/questions/45323619/>`_.

    Tolerance values should be set in relation to the granularity
    (e.g., pre-rounding) of the input data.

    Warning
    -------
        Float GCD calcualtion is inherently approximate.
        Mixing floats with other types in
        `calculate_gcd` will reduce reliability.
        Prefer `Fraction` where possible.
        For solutions specific to music, see other modeules in this repo,
        notably `grid`.

    Parameters
    ----------
    a: float
        Any float value.
    b: float
        Any float value, though typically 1.0 (default) for our use case
        of measure-relative positioning.
    rtol
        the relative tolerance
    atol
        the absolute tolerance

    Examples
    --------
    At risk of failure in both directions.
    Default tolerance values fail simple cases (2 / 3 to 4d.p.):
    >>> round(float_gcd_pair(0.6667), 3) # failure
    0.0

    Leaving the value the same, but changing the tolerance to accommodate:
    >>> round(float_gcd_pair(0.6667, atol=0.001, rtol=0.001), 3) # success
    0.333

    But this same kind of tolerance adjustment can make errors for other,
    common musical values.
    15/16 is a common musical value for which the finer tolerance is effective:

    >>> fifteen_sixteenths = 15/16
    >>> round(1 / float_gcd_pair(fifteen_sixteenths)) # success
    16

    >>> round(1 / float_gcd_pair(fifteen_sixteenths, atol=0.001, rtol=0.001)) # success
    16

    >>> fifteen_sixteenths_3dp = round(fifteen_sixteenths, 3)
    >>> round(1 / float_gcd_pair(fifteen_sixteenths_3dp)) # failure
    500

    >>> round(1 / float_gcd_pair(fifteen_sixteenths_3dp, atol=0.001, rtol=0.001)) # failure
    500

    Note that both-zero inputs return zero:
    >>> float_gcd_pair(0.0, 0.0)
    0.0

    """
    if a == 0.0 and b == 0.0:
        return 0.0
    t = min(abs(a), abs(b))
    while abs(b) > rtol * t + atol:
        a, b = b, a % b
    return a


def lcm_pair(a: int, b: int) -> int:
    """
    Compute the Lowest Common Multiple (LCM) of two integers.

    >>> lcm_pair(8, 16)
    16

    >>> lcm_pair(2, 3)
    6

    """
    return a * b // integer_gcd_pair(a, b)


def fraction_gcd_pair(x: Fraction, y: Fraction) -> Fraction:
    """
    Compute the GCD of two fractions using the
    equivalence between gcd(a/b, c/d) and gcd(a, c)/lcm(b, d)

    This function compares exactly two fractions (x and y).
    For a longer list, use `fraction_gcd`.

    Returns
    -------
    Fraction
        The GCD of `x` and `y`, which is always simplified.

    Examples
    --------
    >>> fraction_gcd_pair(Fraction(1, 2), Fraction(2, 3))
    Fraction(1, 6)

    """
    return Fraction(
        integer_gcd_pair(x.numerator, y.numerator),
        lcm_pair(x.denominator, y.denominator),
    )


# -----------------------------------------------------------------------------

# Combined operations on iterable (list, tuple, ... ) of values.


def calculate_gcd(numbers: Sequence):
    """
    Compute GCD.
    Wrapper function when you don't know the type of the numeric data.
    If the value type is known (integer, fractions, float),
    use the more specific ``{type}_gcd`` function.

    Integers and fractions are lossless and processed first, before any floats.

    Warning
    -------
        Mixing floats with other numeric types reduces reliability.
        Prefer `Fraction` where possible.

    Raises
    ------
    ValueError
        If `numbers` is empty.

    >>> calculate_gcd([1, 2])
    Fraction(1, 1)

    >>> calculate_gcd([1, Fraction(1, 2), 2])
    Fraction(1, 2)

    >>> calculate_gcd([0, 1/2])
    0.5

    >>> calculate_gcd([0, Fraction(1, 2), 1/2])
    Fraction(1, 2)

    >>> gcd = calculate_gcd([0, Fraction(1, 2), 4/12])
    >>> round(gcd, 3)
    0.167

    All-float input is supported:
    >>> calculate_gcd([0.5, 0.25])
    0.25

    """
    if not numbers:
        raise ValueError("numbers must not be empty")

    floats = [num for num in numbers if isinstance(num, float)]
    ints_fractions = [num for num in numbers if not isinstance(num, float)]

    if ints_fractions:
        gcd = Fraction(ints_fractions[0])
        for num in ints_fractions[1:]:
            gcd = fraction_gcd_pair(Fraction(num), gcd)
        for f in floats:
            gcd = float_gcd_pair(f, gcd)
    else:
        # All floats
        gcd = floats[0]
        for f in floats[1:]:
            gcd = float_gcd_pair(f, gcd)

    return gcd


def integer_gcd(integers: Sequence[int]) -> int:
    """
    Compute GCD where the elements are known/asserted to be integers.
    See `integer_gcd_pair`.

    Raises
    ------
    ValueError
        If `integers` is empty.
        Note that this diverges from `math.gcd`.

    Returns
    -------
    int
        The GCD of all elements in the list.

    Examples
    --------
    >>> integer_gcd([0, 2, 4])
    2

    >>> integer_gcd([0, 15, 16])
    1

    >>> integer_gcd([0, 8, 16])
    8

    Zero returns 0 ...
    >>> integer_gcd([0])
    0

    ... but nothing raises an error.
    >>> integer_gcd([])
    Traceback (most recent call last):
    ValueError: integers must not be empty

    """
    if not integers:
        raise ValueError("integers must not be empty")
    gcd = integers[0]
    for i in range(1, len(integers)):
        gcd = integer_gcd_pair(gcd, integers[i])
    return gcd


def fraction_gcd(fractions: Sequence[Fraction]) -> Fraction:
    """
    Compute GCD where all elements are known/asserted to be Fractions.
    See `fraction_gcd_pair`.

    Raises
    ------
    ValueError
        If `fractions` is empty.

    Returns
    -------
    Fraction
        The GCD of all Fractions in `fractions`.

    Examples
    --------
    >>> fraction_gcd([Fraction(1, 2), Fraction(2, 3), Fraction(5, 12)])
    Fraction(1, 12)

    """
    if not fractions:
        raise ValueError("fractions must not be empty")
    gcd = fractions[0]
    for i in range(1, len(fractions)):
        gcd = fraction_gcd_pair(gcd, fractions[i])
    return gcd


def float_gcd(floats: Sequence[float], rtol=1e-05, atol=1e-08) -> float:
    """
    Calculate GCD for a list of floats given the specified
    relative and absolute tolerance (`rtol` and `atol`).

    If the values are known to be integers use `integer_gcd`, known to be
    Fractions use `fraction_gcd`, and if the type is not known use
    `calculate_gcd`.

    Warning
    -------
        Float GCD is inherently approximate.
        See `float_gcd_pair` for details.

    Raises
    ------
    ValueError
        If `floats` is empty.

    Parameters
    ----------
    floats: list[float]
        Any float values.
    rtol
        the relative tolerance
    atol
        the absolute tolerance

    """
    if not floats:
        raise ValueError("floats must not be empty")
    gcd = floats[0]
    for i in range(1, len(floats)):
        gcd = float_gcd_pair(gcd, floats[i], rtol=rtol, atol=atol)
    return gcd


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from typing import List

import pretty_midi
import pytest

from amads.core.basics import Note, Score
from amads.io.readscore import (
    last_used_reader,
    read_score,
    set_preferred_midi_reader,
)
from amads.music import example

VERBOSE = False  # set to True for more debug output


def basic_note_compare(
    score_notes: List[Note], pm_notes: List[pretty_midi.Note]
):
    """Compare a list of AMADS Notes with a list of pretty_midi Notes."""
    for i in range(min(len(score_notes), len(pm_notes))):
        score_note = score_notes[i]
        pm_note = pm_notes[i]
        if VERBOSE:
            print(
                f"Comparing note {i}: AMADS onset {score_note.onset} "
                f"pitch {score_note.key_num} duration {score_note.duration} | "
                f"PM onset {pm_note.start} pitch {pm_note.pitch} "
                f"duration {pm_note.end - pm_note.start}"
            )
        if (
            score_note.onset != pytest.approx(pm_note.start, abs=1e-3)
            or score_note.duration
            != pytest.approx(pm_note.end - pm_note.start, abs=1e-3)
            or score_note.key_num != pm_note.pitch
        ):
            print(f"NOTE MISMATCH IN TEST at index {i}")
            print(
                f"AMADS note: onset {score_note.onset} pitch {score_note.key_num} "
                f"duration {score_note.duration}"
            )
            print(
                f"PM note: onset {pm_note.start} pitch {pm_note.pitch} "
                f"duration {pm_note.end - pm_note.start}"
            )
            # now add assertions so debug output will tell us what failed
            assert score_note.onset == pytest.approx(
                pm_note.start, abs=1e-3
            ), f"Onset mismatch at note {i}"

            # duration testing is pretty lax because with partitura, we try to avoid
            # ties to very short notes, and that can change the duration by more than
            # just rounding error.
            assert score_note.duration == pytest.approx(
                pm_note.end - pm_note.start, abs=1e-3
            ), f"Duration mismatch at note {i}"
            assert (
                score_note.key_num == pm_note.pitch
            ), f"Pitch mismatch at note {i}"
    assert len(score_notes) == len(pm_notes), (
        f"Number of notes mismatch: AMADS has {len(score_notes)} notes, "
        f"pretty_midi has {len(pm_notes)} notes."
    )


@pytest.mark.parametrize(
    "midi_filename",
    [
        "tempochange.mid",
        "sarabande.mid",
        # "chopin_prelude_7.mid",  # chopin_prelude_7.mid has overlapping notes
        # on same pitch, which is no well-defined
        "twochan.mid",
    ],
)
def test_import_midi(midi_filename):
    """
    Test MIDI import by comparing the results with pretty_midi.

    Parameters
    ----------
    midi_filename : str
        Name of the MIDI file to test
    """
    midi_file = example.fullpath(f"midi/{midi_filename}")
    assert midi_file is not None
    set_preferred_midi_reader(reader="pretty_midi")
    pmscore = read_score(midi_file, show=VERBOSE)
    assert isinstance(pmscore, Score)
    pmscore = pmscore.flatten(collapse=True)
    print(
        f"Imported MIDI file {midi_filename} into AMADS Score "
        f"using {last_used_reader()}."
    )
    # pmscore.quantize(12)
    # pmscore = pmscore.merge_tied_notes()  # so we can count notes from MIDI correctly
    # score_notes = pmscore.list_all(Note)

    # print("AMADS notes:")
    # for note in score_notes:
    #     print(f"{note.onset / 4:0.2f} {note.key_num} {note.duration:0.2f}")

    pmscore.convert_to_seconds()
    if VERBOSE:
        print("PrettyMIDI imported score converted to SECONDS:")
        pmscore.show()
    pmscore_notes = pmscore.get_sorted_notes()
    for note in pmscore_notes:
        assert note.dynamic != 0, "Note has zero dynamic after import."

    pm = pretty_midi.PrettyMIDI(str(midi_file))
    print(f"PrettyMIDI resolution: {pm.resolution}")
    pm_notes = [
        note for instrument in pm.instruments for note in instrument.notes
    ]

    # print("PrettyMIDI notes before sort:")
    # for note in pm_notes:
    #     print(f"{note.start:0.2f} {note.pitch} {note.end - note.start:0.2f}")

    pm_notes.sort(key=lambda x: (x.start, x.pitch))
    print("PrettyMIDI notes after sort:")
    for note in pm_notes:
        print(f"{note.start:0.2f} {note.pitch} {note.end - note.start:0.2f}")

    basic_note_compare(pmscore_notes, pm_notes)

    set_preferred_midi_reader(reader="music21")
    m2score = read_score(midi_file, show=VERBOSE)
    m2score.convert_to_seconds()
    assert isinstance(m2score, Score)
    print(
        f"Imported MIDI file {midi_filename} into AMADS Score "
        f"using {last_used_reader()}."
    )
    print("AMADS notes from music21 MIDI import:")
    m2score.show()
    basic_note_compare(m2score.get_sorted_notes(), pm_notes)
